"""Univariate Distribution module."""

from skfolio.distribution.univariate._base import BaseUnivariateDist
from skfolio.distribution.univariate._gaussian import Gaussian
from skfolio.distribution.univariate._johnson_su import JohnsonSU
from skfolio.distribution.univariate._normal_inverse_gaussian import (
    NormalInverseGaussian,
)
from skfolio.distribution.univariate._selection import select_univariate_dist
from skfolio.distribution.univariate._student_t import StudentT

__all__ = [
    "BaseUnivariateDist",
    "Gaussian",
    "JohnsonSU",
    "NormalInverseGaussian",
    "StudentT",
    "select_univariate_dist",
]



"""Graphical Lasso CV Covariance Estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import sklearn.covariance as skc

from skfolio.moments.covariance._base import BaseCovariance


class GraphicalLassoCV(BaseCovariance, skc.GraphicalLassoCV):
    """Sparse inverse covariance with cross-validated choice of the l1 penalty.

    Read more in `scikit-learn
    <https://scikit-learn.org/stable/auto_examples/covariance/plot_sparse_cov.html>`_.

    Parameters
    ----------
    alphas : int or array-like of shape (n_alphas,), dtype=float, default=4
        If an integer is given, it fixes the number of points on the
        grids of alpha to be used. If a list is given, it gives the
        grid to be used. See the notes in the class docstring for
        more details. Range is [1, inf) for an integer.
        Range is (0, inf] for an array-like of floats.

    n_refinements : int, default=4
        The number of times the grid is refined. Not used if explicit
        values of alphas are passed. Range is [1, inf).

    cv : int, cross-validation generator or iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - `CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs :class:`KFold` is used.

    tol : float, default=1e-4
        The tolerance to declare convergence: if the dual gap goes below
        this value, iterations are stopped. Range is (0, inf].

    enet_tol : float, default=1e-4
        The tolerance for the elastic net solver used to calculate the descent
        direction. This parameter controls the accuracy of the search direction
        for a given column update, not of the overall parameter estimate. Only
        used for mode='cd'. Range is (0, inf].

    max_iter : int, default=100
        Maximum number of iterations.

    mode : {'cd', 'lars'}, default='cd'
        The Lasso solver to use: coordinate descent or LARS. Use LARS for
        very sparse underlying graphs, where number of features is greater
        than number of samples. Elsewhere prefer cd which is more numerically
        stable.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        `None` means 1 unless in a :obj:`joblib.parallel_backend` context.
        `-1` means using all processors.

    verbose : bool, default=False
        If verbose is True, the objective function and duality gap are
        printed at each iteration.

    assume_centered : bool, default=False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    Attributes
    ----------
    covariance_ : ndarray of shape (n_assets, n_assets)
        Estimated covariance.

    location_ : ndarray of shape (n_assets,)
        Estimated location, i.e. the estimated mean.

    precision_ : ndarray of shape (n_assets, n_assets)
        Estimated pseudo inverse matrix.
        (stored only if store_precision is True)

    alpha_ : float
        Penalization parameter selected.

    cv_results_ : dict of ndarrays
        A dict with keys:

        alphas : ndarray of shape (n_alphas,)
            All penalization parameters explored.

        split(k)_test_score : ndarray of shape (n_alphas,)
            Log-likelihood score on left-out data across (k)th fold.

            .. versionadded:: 1.0

        mean_test_score : ndarray of shape (n_alphas,)
            Mean of scores over the folds.

            .. versionadded:: 1.0

        std_test_score : ndarray of shape (n_alphas,)
            Standard deviation of scores over the folds.

            .. versionadded:: 1.0

    n_iter_ : int
        Number of iterations run for the optimal alpha.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Notes
    -----
    The search for the optimal penalization parameter (`alpha`) is done on an
    iteratively refined grid: first the cross-validated scores on a grid are
    computed, then a new refined grid is centered around the maximum, and so
    on.

    One of the challenges which is faced here is that the solvers can
    fail to converge to a well-conditioned estimate. The corresponding
    values of `alpha` then come out as missing values, but the optimum may
    be close to these missing values.

    In `fit`, once the best parameter `alpha` is found through
    cross-validation, the model is fit again using the entire training set.
    """

    def __init__(
        self,
        alphas=4,
        n_refinements=4,
        cv=None,
        tol=1e-4,
        enet_tol=1e-4,
        max_iter=100,
        mode="cd",
        n_jobs=None,
        verbose=False,
        assume_centered=False,
        nearest: bool = True,
        higham: bool = False,
        higham_max_iteration: int = 100,
    ):
        super().__init__(
            nearest=nearest,
            higham=higham,
            higham_max_iteration=higham_max_iteration,
        )
        skc.GraphicalLassoCV.__init__(
            self,
            alphas=alphas,
            n_refinements=n_refinements,
            cv=cv,
            tol=tol,
            enet_tol=enet_tol,
            max_iter=max_iter,
            mode=mode,
            n_jobs=n_jobs,
            verbose=verbose,
            assume_centered=assume_centered,
        )

    def fit(self, X, y=None, **fit_params) -> "GraphicalLassoCV":
        """Fit the GraphicalLasso covariance model to X.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
         Price returns of the assets.

        y : Ignored
          Not used, present for API consistency by convention.

        Returns
        -------
        self : GraphicalLassoCV
          Fitted estimator.
        """
        skc.GraphicalLassoCV.fit(self, X, **fit_params)
        self._set_covariance(self.covariance_)
        return self



"""Bivariate Joe Copula Estimation."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import scipy.optimize as so
import scipy.special as sp
import scipy.stats as st
import sklearn.utils.validation as skv

from skfolio.distribution.copula._base import BaseBivariateCopula
from skfolio.distribution.copula._utils import (
    CopulaRotation,
    _apply_copula_rotation,
    _apply_margin_swap,
    _apply_rotation_cdf,
    _apply_rotation_partial_derivatives,
    _select_rotation_itau,
    _select_theta_and_rotation_mle,
)

# Joe copula with a theta of 1.0 is just the independence copula, so we chose a lower
# bound of 1.005. After 20, the copula is already imposing very high tail dependence
# closed to comonotonic and increasing it will make it impractical.
_THETA_BOUNDS = (1.005, 20.0)
_EULER_GAMMA = 0.5772156649015328606


class JoeCopula(BaseBivariateCopula):
    r"""Bivariate Joe Copula Estimation.

    The Joe copula is an Archimedean copula characterized by strong upper tail
    dependence and little to no lower tail dependence.

    In its unrotated form, it is used for modeling extreme co-movements in the upper
    tail (i.e. simultaneous extreme gains).

    Rotations allow the copula to be adapted for different types of tail dependence:
      - A 180° rotation captures extreme co-movements in the lower tail (i.e.
        simultaneous extreme losses).

      - A 90° rotation captures scenarios where one variable exhibits extreme losses
        while the other shows extreme gains.

      - A 270° rotation captures the opposite scenario, where one variable experiences
        extreme gains while the other suffers extreme losses.

    Joe copula generally exhibits stronger upper tail dependence than the Gumbel copula.

    It is defined by:

    .. math::
            C_{\theta}(u, v) = 1-\Bigl[(1 - u)^{\theta} + (1 - v)^{\theta} -
                (1 - u)^{\theta} (1 - v)^{\theta}\Bigr]^{\frac{1}{\theta}}

    where :math:`\theta \ge 1` is the dependence parameter. When :math:`\theta = 1`,
    the Joe copula reduces to the independence copula. Larger values of :math:`\theta`
    result in stronger upper-tail dependence.

    .. note::

        Rotation are needed for archimedean copulas (e.g., Joe, Gumbel, Clayton)
        because their parameters only model positive dependence, and they exhibit
        asymmetric tail behavior. To model negative dependence, one uses rotations
        to “flip” the copula's tail dependence.

    Parameters
    ----------
    itau : bool, default=True
        If True, :math:`\theta` is estimated using the Kendall's tau inversion method;
        otherwise, the Maximum Likelihood Estimation (MLE) method is used. The MLE is
        slower but more accurate.

    kendall_tau : float, optional
        If `itau` is True and `kendall_tau` is provided, this value is used;
        otherwise, it is computed.

    tolerance : float, default=1e-4
        Convergence tolerance for the MLE optimization.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    theta_ : float
        Fitted theta coefficient :math:`\theta` > 1.

    rotation_ : CopulaRotation
        Fitted rotation of the copula.

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution import JoeCopula, compute_pseudo_observations
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_dataset()
    >>> X = prices_to_returns(prices)
    >>> X = X[["AAPL", "JPM"]]
    >>>
    >>> # Convert returns to pseudo observation in the interval [0,1]
    >>> X = compute_pseudo_observations(X)
    >>>
    >>> # Initialize the Copula estimator
    >>> model = JoeCopula()
    >>>
    >>> # Fit the model to the data.
    >>> model.fit(X)
    >>>
    >>> # Display the fitted parameter and tail dependence coefficients
    >>> print(model.fitted_repr)
    JoeCopula(theta=1.48, rot=180°)
    >>> print(model.lower_tail_dependence)
    0.4021
    >>> print(model.upper_tail_dependence)
    0.0
    >>>
    >>> # Compute the log-likelihood, total log-likelihood, CDF, Partial Derivative,
    >>> # Inverse Partial Derivative, AIC, and BIC
    >>> log_likelihood = model.score_samples(X)
    >>> score = model.score(X)
    >>> cdf = model.cdf(X)
    >>> p = model.partial_derivative(X)
    >>> u = model.inverse_partial_derivative(X)
    >>> aic = model.aic(X)
    >>> bic = model.bic(X)
    >>>
    >>> # Generate 5 new samples
    >>> samples = model.sample(n_samples=5)
    >>>
    >>> # Plot the tail concentration function.
    >>> fig = model.plot_tail_concentration()
    >>> fig.show()
    >>>
    >>> # Plot a 2D contour of the estimated PDF.
    >>> fig = model.plot_pdf_2d()
    >>> fig.show()
    >>>
    >>> # Plot a 3D surface of the estimated PDF.
    >>> fig = model.plot_pdf_3d()
    >>> fig.show()

    References
    ----------
    .. [1] "An Introduction to Copulas (2nd ed.)",
       Nelsen (2006)

    .. [2] "Multivariate Models and Dependence Concepts",
        Joe, Chapman & Hall (1997)

    .. [3] "Quantitative Risk Management: Concepts, Techniques and Tools",
        McNeil, Frey & Embrechts (2005)

    .. [4] "The t Copula and Related Copulas",
        Demarta & McNeil (2005)

    .. [5] "Copula Methods in Finance",
        Cherubini, Luciano & Vecchiato (2004)
    """

    theta_: float
    rotation_: CopulaRotation
    _n_params = 1

    def __init__(
        self,
        itau: bool = True,
        kendall_tau: float | None = None,
        tolerance: float = 1e-4,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.itau = itau
        self.kendall_tau = kendall_tau
        self.tolerance = tolerance

    def fit(self, X: npt.ArrayLike, y=None) -> "JoeCopula":
        r"""Fit the Bivariate Joe Copula.

        If `itau` is True, estimates :math:`\theta` using Kendall's tau inversion.
        Otherwise, uses MLE by maximizing the log-likelihood.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        X = self._validate_X(X, reset=True)

        if self.itau:
            if self.kendall_tau is None:
                kendall_tau = st.kendalltau(X[:, 0], X[:, 1]).statistic
            else:
                kendall_tau = self.kendall_tau

            abs_kendall_tau = abs(kendall_tau)

            # Root-finding function brentq to find the value of theta in the interval
            # brentq fails if _tau_diff has same sign, it happens when we are at the
            # bounds so we capture it before.
            fa = _tau_diff(_THETA_BOUNDS[0], abs_kendall_tau)
            fb = _tau_diff(_THETA_BOUNDS[1], abs_kendall_tau)
            if fa * fb > 0:
                if abs(fa) < abs(fb):
                    self.theta_ = _THETA_BOUNDS[0]
                else:
                    self.theta_ = _THETA_BOUNDS[1]
            else:
                # noinspection PyTypeChecker
                self.theta_ = so.brentq(
                    _tau_diff,
                    args=(abs_kendall_tau,),
                    a=_THETA_BOUNDS[0],
                    b=_THETA_BOUNDS[-1],
                )
            self.rotation_ = _select_rotation_itau(
                func=_neg_log_likelihood, X=X, theta=self.theta_
            )

        else:
            self.theta_, self.rotation_ = _select_theta_and_rotation_mle(
                _neg_log_likelihood, X=X, bounds=_THETA_BOUNDS, tolerance=self.tolerance
            )

        return self

    def cdf(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the CDF of the bivariate Joe copula.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        Returns
        -------
        cdf : ndarray of shape (n_observations,)
            CDF values for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        cdf = _apply_rotation_cdf(
            func=_base_cdf, X=X, rotation=self.rotation_, theta=self.theta_
        )
        return cdf

    def partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the h-function (partial derivative) for the bivariate Joe copula
        with respect to a specified margin.

        The h-function with respect to the second margin represents the conditional
        distribution function of :math:`u` given :math:`v`:

        .. math::  \begin{aligned}
                   h(u \mid v)
                     &= \frac{\partial C(u,v)}{\partial v} \\[6pt]
                     &= (1-v)^{\theta-1}\,\Bigl[1 \;-\;(1-u)^{\theta}\Bigr]\,
                        \Bigl[(1-u)^{\theta} \;+\;(1-v)^{\theta}
                              \;-\;(1-u)^{\theta}(1-v)^{\theta}\Bigr]^{\frac{1}{\theta}-1} \\[6pt]
                     &= \left( 1 \;+\;\frac{(1-u)^{\theta}}{(1-v)^{\theta}}
                              \;-\;(1-u)^{\theta} \right)^{-1 + \frac{1}{\theta}}
                        \;\cdot\;\bigl[\,1 \;-\;(1-u)^{\theta}\bigr].
                   \end{aligned}

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        first_margin : bool, default=False
            If True, compute the partial derivative with respect to the first
            margin `u`; otherwise, compute the partial derivative with respect to the
            second margin `v`.

        Returns
        -------
        p  : ndarray of shape (n_observations,)
            h-function values :math:`h(u \mid v) \;=\; p` for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        p = _apply_rotation_partial_derivatives(
            func=_base_partial_derivative,
            X=X,
            rotation=self.rotation_,
            first_margin=first_margin,
            theta=self.theta_,
        )
        return p

    def inverse_partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the inverse of the bivariate copula's partial derivative, commonly
        known as the inverse h-function [1]_.

        Let :math:`C(u, v)` be a bivariate copula. The h-function with respect to the
        second margin is defined by

        .. math::
            h(u \mid v) \;=\; \frac{\partial\,C(u, v)}{\partial\,v},

        which is the conditional distribution of :math:`U` given :math:`V = v`.
        The **inverse h-function**, denoted :math:`h^{-1}(p \mid v)`, is the unique
        value :math:`u \in [0,1]` such that

        .. math::
            h(u \mid v) \;=\; p,
            \quad \text{where } p \in [0,1].

        In practical terms, given :math:`(p, v)` in :math:`[0, 1]^2`,
        :math:`h^{-1}(p \mid v)` solves for the :math:`u` satisfying
        :math:`p = \partial C(u, v)/\partial v`.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(p, v)`, each in the interval `[0, 1]`.
            - The first column `p` corresponds to the value of the h-function.
            - The second column `v` is the conditioning variable.

        first_margin : bool, default=False
            If True, compute the inverse partial derivative with respect to the first
            margin `u`; otherwise, compute the inverse partial derivative with respect
            to the second margin `v`.

        Returns
        -------
        u : ndarray of shape (n_observations,)
            A 1D-array of length `n_observations`, where each element is the computed
            :math:`u = h^{-1}(p \mid v)` for the corresponding pair in `X`.

        References
        ----------
        .. [1] "Multivariate Models and Dependence Concepts", Joe, H. (1997)
        .. [2] "An Introduction to Copulas", Nelsen, R. B. (2006)
        .. [3] . "Nested Archimedean Copulas Meet ", Hofert & Mächler (2011)
        """
        # no known closed-form solution, hence we use Newton method.
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        u = _apply_rotation_partial_derivatives(
            func=_base_inverse_partial_derivative,
            X=X,
            rotation=self.rotation_,
            first_margin=first_margin,
            theta=self.theta_,
        )
        return u

    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the log-likelihood of each sample (log-pdf) under the model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            The log-likelihood of each sample under the fitted copula.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        X = _apply_copula_rotation(X, rotation=self.rotation_)
        log_density = _base_sample_scores(X=X, theta=self.theta_)
        return log_density

    @property
    def lower_tail_dependence(self) -> float:
        """Theoretical lower tail dependence coefficient."""
        skv.check_is_fitted(self)
        if self.rotation_ == CopulaRotation.R180:
            return 2.0 - np.power(2.0, 1.0 / self.theta_)
        return 0

    @property
    def upper_tail_dependence(self) -> float:
        """Theoretical upper tail dependence coefficient."""
        skv.check_is_fitted(self)
        if self.rotation_ == CopulaRotation.R0:
            return 2.0 - np.power(2.0, 1.0 / self.theta_)
        return 0

    @property
    def fitted_repr(self) -> str:
        """String representation of the fitted copula."""
        return (
            f"{self.__class__.__name__}(theta={self.theta_:0.2f}, rot={self.rotation_})"
        )


def _neg_log_likelihood(theta: float, X: np.ndarray) -> float:
    """Negative log-likelihood function for optimization.

    Parameters
    ----------
     X : array-like of shape (n_observations, 2)
         An array of bivariate inputs `(u, v)` where each row represents a
         bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
         having been transformed to uniform marginals.

    theta : float
         The dependence parameter (must be greater than 1).

    Returns
    -------
     value : float
         The negative log-likelihood value.
    """
    return -np.sum(_base_sample_scores(X=X, theta=theta))


def _base_sample_scores(X: np.ndarray, theta: float) -> np.ndarray:
    """Compute the log-likelihood of each sample (log-pdf) under the bivariate
    Joe copula model.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs `(u, v)` where each row represents a
        bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
        having been transformed to uniform marginals.

    theta : float
        The dependence parameter (must be greater than 1).

    Returns
    -------
    density : ndarray of shape (n_observations,)
        The log-likelihood of each sample under the fitted copula.

    Raises
    ------
    ValueError
        If rho is not in (-1, 1) or dof is not positive.
    """
    if theta <= 1.0:
        raise ValueError("Theta must be greater than 1 for the Joe copula.")

    # log-space transformation to improve stability near 0  or 1
    x, y = np.log1p(-X).T
    x_y = x + y
    d = np.exp(x * theta) + np.exp(y * theta) - np.exp(x_y * theta)
    log_density = (
        (1.0 / theta - 2.0) * np.log(d) + x_y * (theta - 1.0) + np.log(theta - 1.0 + d)
    )
    return log_density


def _tau_diff(theta: float, tau_empirical: float) -> float:
    r"""Compute the difference between the theoretical Kendall's tau for the Joe copula
    and an empirical tau.

    The theoretical relationship for the Joe copula is given by:

    .. math::
       \tau(\theta) = 1 + \frac{2}{2-\theta} \left[ (1-\gamma) - \psi\left(\frac{2}{\theta}+1\right) \right],

    where :math:`\psi` is the digamma function and :math:`\gamma` is the
    Euler-Mascheroni constant.

    Parameters
    ----------
    theta : float
        The dependence parameter (must be greater than 1).

    tau_empirical : float
        The empirical Kendall's tau.

    Returns
    -------
    float
        The difference :math:`\tau(\theta) - \tau_{\text{empirical}}`.
    """
    # Euler-Mascheroni constant: gamma_const = 1 - EulerGamma
    gamma_const = 1.0 - _EULER_GAMMA
    # Compute theoretical tau using the digamma-based expression
    tau_theoretical = 1.0 + (2.0 / (2.0 - theta)) * (
        gamma_const - sp.digamma(2.0 / theta + 1.0)
    )
    return tau_theoretical - tau_empirical


def _base_cdf(X: np.ndarray, theta: float) -> np.ndarray:
    """Bivariate Joe CDF (unrotated)."""
    z = np.power(1 - X, theta)
    cdf = 1.0 - np.power(np.sum(z, axis=1) - np.prod(z, axis=1), 1.0 / theta)
    return cdf


def _base_partial_derivative(
    X: np.ndarray, first_margin: bool, theta: float
) -> np.ndarray:
    r"""Compute the h-function (partial derivative) for the bivariate unrotated
    Joe copula with respect to a specified margin.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs `(u, v)` where each row represents a
        bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
        having been transformed to uniform marginals.

    first_margin : bool, default=False
        If True, compute the partial derivative with respect to the first
        margin `u`; otherwise, compute the partial derivative with respect to the
        second margin `v`.

    theta : float
        The dependence parameter (must be greater than 1).

    Returns
    -------
      : ndarray of shape (n_observations,)
        h-function values :math:`h(u \mid v) \;=\; p` for each observation in X.
    """
    X = _apply_margin_swap(X, first_margin=first_margin)
    x, y = np.power(1 - X, theta).T
    p = np.power(1 + x / y - x, 1 / theta - 1) * (1.0 - x)
    return p


def _base_inverse_partial_derivative(
    X: np.ndarray, first_margin: bool, theta: float
) -> np.ndarray:
    r"""Compute the inverse of the bivariate copula's partial derivative, commonly
    known as the inverse h-function.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs `(p, v)`, each in the interval `[0, 1]`.
        - The first column `p` corresponds to the value of the h-function.
        - The second column `v` is the conditioning variable.

    first_margin : bool, default=False
        If True, compute the inverse partial derivative with respect to the first
        margin `u`; otherwise, compute the inverse partial derivative with respect to
        the second margin `v`.

    theta : float
        The dependence parameter (must be greater than 1).

    Returns
    -------
    u : ndarray of shape (n_observations,)
        A 1D-array of length `n_observations`, where each element is the computed
        :math:`u = h^{-1}(p \mid v)` for the corresponding pair in `X`.
    """
    X = _apply_margin_swap(X, first_margin=first_margin)

    p, v = X.T

    y = np.power(1 - v, theta)

    # No known closed-form solution, hence we use Newton method
    # with an early-stopping criterion

    # Initial guess
    x = np.power(
        (1 - v) * (np.power(1.0 - p, 1.0 / theta - 1) - 1.0) / y + 1.0,
        theta / (1.0 - theta),
    )

    max_iters = 50
    tol = 1e-8
    for _ in range(max_iters):
        k = (x - 1.0) * y
        w = np.power((1.0 / y - 1.0) * x + 1.0, 1.0 / theta)
        x_new = (
            x
            - (theta * (k - x) * (p * (-k + x) + k * w))
            / ((y - 1.0) * k - theta * y)
            / w
        )
        x_new = np.clip(x_new, 0.0, 1.0)
        diff = np.max(np.abs(x_new - x))
        x = x_new
        if diff < tol:
            break

    u = 1.0 - np.power(x, 1.0 / theta)
    return u



"""Multiple Randomized CV."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credit: Daniel Palomar
# SPDX-License-Identifier: BSD-3-Clause

import math
from collections.abc import Iterator
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
import sklearn.utils as sku

from skfolio.model_selection._walk_forward import WalkForward
from skfolio.utils.stats import sample_unique_subsets
from skfolio.utils.tools import safe_split


class MultipleRandomizedCV:
    r"""Multiple Randomized Cross-Validation.

    Based on the "Multiple Randomized Backtests" methodology of Palomar [1]_,
    this cross-validation strategy performs a resampling-based evaluation by repeatedly
    sampling **distinct** asset subsets (without replacement) and **contiguous** time
    windows, then applying an inner walk-forward split to each subsample, capturing both
    temporal and cross-sectional variability in performance.

    On each of the `n_subsamples` iterations, the following actions are performed:

    1. Randomly pick a contiguous time window of length `window_size` (or the full history if None).
    2. Randomly pick an asset subset of size `asset_subset_size` (without replacement).
    3. Run a walk-forward split (via the supplied `walk_forward` object) on that sub-dataset.
    4. Yield `(train_indices, test_indices, asset_indices)` for each inner split.

    Each asset subset is sampled without replacement (assets within each subset are
    distinct) and no subset is repeated across the `n_subsamples` draws. We employ the
    combinatorial unranking algorithm to compute any k-combination in
    `O(n_subsamples * asset_subset_size)` time and space, without generating or storing
    all :math:`M=\binom{n\_assets}{asset\_subset\_size}` subsets. When :math:`M` is
    small, this guarantees exhaustive coverage of every possible asset-universe.
    Because ranks are drawn without replacement from a finite population of size
    :math:`M`, the variance of the sample mean is reduced by the finite-population
    correction factor :math:`\tfrac{M - n\_subsamples}{M - 1}`.

    Parameters
    ----------
    walk_forward : WalkForward
        A :class:`~skfolio.model_selection.WalkForward` CV object to be applied to
        each subsample.

    n_subsamples : int
        Number of independent subsamples (sub-datasets) to draw. Each subsample is a
        (time window x asset subset) on which you run the inner walk-forward.

    asset_subset_size : int
        How many assets to include in each subsample.
        Must be less or equal to the total number of assets.

    window_size : int or None, default=None
        Length of the contiguous time slice (number of observations) for each
        subsample. If None, uses the full time series observations in every draw.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    References
    ----------
    .. [1]  "Portfolio Optimization: Theory and Application", Chapter 8,
            Daniel P. Palomar (2025)

    Examples
    --------
    Tutorials using `MultipleRandomizedCV`:
        * :ref:`sphx_glr_auto_examples_model_selection_plot_1_multiple_randomized_cv.py`


    >>> import numpy as np
    >>> from skfolio.datasets import load_sp500_dataset, load_factors_dataset
    >>> from skfolio.model_selection import WalkForward, MultipleRandomizedCV
    >>> from skfolio.preprocessing import prices_to_returns
    >>>
    >>> X = np.random.randn(4, 5) # 4 observations and 5 assets.
    >>> # Draw 2 subsamples (sub-datasets) with 3 assets chosen randomly among the 5.
    >>> # For each subsample, run a Walk Forward.
    >>> # Use the full time series (no time resampling).
    >>> cv = MultipleRandomizedCV(
    ...     walk_forward=WalkForward(test_size=1, train_size=2),
    ...     n_subsamples=2,
    ...     asset_subset_size=3,
    ...     window_size=None,
    ...     random_state=0,
    ... )
    >>> for i, (train_index, test_index, assets) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train:  index={train_index}")
    ...     print(f"  Test:   index={test_index}")
    ...     print(f"  Assets: columns={assets}")
    Fold 0:
      Train:  index=[0 1]
      Test:   index=[2]
      Assets: columns=[0 1 4]
    Fold 1:
      Train:  index=[1 2]
      Test:   index=[3]
      Assets: columns=[0 1 4]
    Fold 2:
      Train:  index=[0 1]
      Test:   index=[2]
      Assets: columns=[1 3 4]
    Fold 3:
      Train:  index=[1 2]
      Test:   index=[3]
      Assets: columns=[1 3 4]
    >>> print(f"Path ids: {cv.get_path_ids()}")
    Path ids: [0 0 1 1]
    >>>
    >>> # Random contiguous time slice of 4 observations among 10 observations.
    >>> X = np.random.randn(10, 5) # 10 observations and 5 assets.
    >>> cv = MultipleRandomizedCV(
    ...     walk_forward=WalkForward(test_size=1, train_size=2),
    ...     n_subsamples=2,
    ...     asset_subset_size=3,
    ...     window_size=4,
    ...     random_state=0,
    ... )
    >>> for i, (train_index, test_index, assets) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train:  index={train_index}")
    ...     print(f"  Test:   index={test_index}")
    ...     print(f"  Assets: columns={assets}")
    Fold 0:
      Train:  index=[4 5]
      Test:   index=[6]
      Assets: columns=[0 1 4]
    Fold 1:
      Train:  index=[5 6]
      Test:   index=[7]
      Assets: columns=[0 1 4]
    Fold 2:
      Train:  index=[5 6]
      Test:   index=[7]
      Assets: columns=[1 3 4]
    Fold 3:
      Train:  index=[6 7]
      Test:   index=[8]
      Assets: columns=[1 3 4]
    >>>
    >>> # Walk Forward with time-based (calendar) rebalancing.
    >>> # Rebalance every 3 months on the third Friday, and train on the last 12 months.
    >>> prices = load_sp500_dataset()
    >>> X = prices_to_returns(prices)
    >>> X = X["2021":"2022"]
    >>> cv = MultipleRandomizedCV(
    ...     walk_forward=WalkForward(test_size=3, train_size=12, freq="WOM-3FRI"),
    ...     n_subsamples=2,
    ...     asset_subset_size=3,
    ...     window_size=None,
    ...     random_state=0,
    ... )
    >>> for i, (train_index, test_index, assets) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train:  size={len(train_index)}")
    ...     print(f"  Test:   size={len(test_index)}")
    ...     print(f"  Assets: columns={assets}")
    Fold 0:
      Train:  size=256
      Test:   size=59
      Assets: columns=[ 9 16 17]
    Fold 1:
      Train:  size=253
      Test:   size=61
      Assets: columns=[ 9 16 17]
    Fold 2:
      Train:  size=251
      Test:   size=69
      Assets: columns=[ 9 16 17]
    Fold 3:
      Train:  size=256
      Test:   size=59
      Assets: columns=[ 7 10 14]
    Fold 4:
      Train:  size=253
      Test:   size=61
      Assets: columns=[ 7 10 14]
    Fold 5:
      Train:  size=251
      Test:   size=69
      Assets: columns=[ 7 10 14]
    >>> print(f"Path ids: {cv.get_path_ids()}")
    [0 0 0 1 1 1]
    """

    if TYPE_CHECKING:
        _path_ids: list[int]

    def __init__(
        self,
        walk_forward: WalkForward,
        n_subsamples: int,
        asset_subset_size: int,
        window_size: int | None = None,
        random_state: int | None = None,
    ):
        self.walk_forward = walk_forward
        self.n_subsamples = n_subsamples
        self.asset_subset_size = asset_subset_size
        self.window_size = window_size
        self.random_state = random_state

    def split(
        self, X: npt.ArrayLike, y=None
    ) -> Iterator[tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_targets)
            Always ignored, exists for compatibility.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.

        assets : ndarray
            The assets indices for that split.
        """
        X, y = sku.indexable(X, y)
        n_observations, n_assets = X.shape
        rng = sku.check_random_state(self.random_state)

        if not isinstance(self.walk_forward, WalkForward):
            raise TypeError("`walk_forward` must be a `WalkForward` instance")

        if self.window_size is not None:
            if not isinstance(self.window_size, int):
                raise ValueError("`window_size` must be an integer")
            if self.window_size < 2 or self.window_size > n_observations:
                raise ValueError(
                    f"When not None, window_size={self.window_size} must "
                    f"satisfy 2 <= window_size <= n_observations={n_observations}."
                )

        if not isinstance(self.asset_subset_size, int):
            raise TypeError("`asset_subset_size` must be an integer")
        if self.asset_subset_size < 1 or self.asset_subset_size > n_assets:
            raise ValueError(
                f"asset_subset_size={self.asset_subset_size} must satisfy "
                f"1 <= asset_subset_size <= n_assets={n_assets}."
            )

        if not isinstance(self.n_subsamples, int):
            raise TypeError("`num_subsamples` must be an integer")

        n_comb = math.comb(n_assets, self.asset_subset_size)
        if self.n_subsamples < 2 or self.n_subsamples > n_comb:
            raise ValueError(
                f"n_subsample={self.n_subsamples} must satisfy 2 <= n_subsample <= "
                f"C({n_assets},{self.asset_subset_size})={n_comb}."
            )

        if n_comb < 1e9:
            asset_indices = sample_unique_subsets(
                n=n_assets,
                k=self.asset_subset_size,
                n_subsets=self.n_subsamples,
                random_state=self.random_state,
            )
        else:
            # Avoid overflow when n_comb is huge (risk of subset duplication is very
            # small and statistically imaterial).
            asset_indices = np.array(
                [
                    rng.choice(n_assets, size=self.asset_subset_size, replace=False)
                    for _ in range(self.n_subsamples)
                ]
            )
            asset_indices.sort(axis=1)

        self._path_ids = []
        for i in range(self.n_subsamples):
            if self.window_size is None:
                start_obs = 0
                X_sample = X
            else:
                start_obs = rng.randint(low=0, high=n_observations - self.window_size)
                obs_indices = np.arange(start_obs, start_obs + self.window_size)
                X_sample, _ = safe_split(X, indices=obs_indices, axis=0)

            for train_indices, test_indices in self.walk_forward.split(X_sample):
                self._path_ids.append(i)
                yield (
                    train_indices + start_obs,
                    test_indices + start_obs,
                    asset_indices[i, :],
                )

    def get_path_ids(self) -> np.ndarray:
        """Return the path id of each test sets in each split."""
        if not hasattr(self, "_path_ids"):
            raise ValueError("Before calling `get_path_ids()` you must call `split(X)`")
        return np.array(self._path_ids)



"""FailedPortfolio sentinel class."""


# Copyright (c) 2025
# Author: Hugo Delatte <hugo.delatte@skfoliolabs.com>
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt

import skfolio.typing as skt
from skfolio.portfolio._portfolio import Portfolio


class FailedPortfolio(Portfolio):
    r"""
    Portfolio object returned when an optimization step fails. It acts as a sentinel
    value that marks the failure and stores failure diagnostics (`optimization_error`,
    `fallback_chain`).

    `FailedPortfolio` preserves full API compatibility with `Portfolio` so it
    can seamlessly pass through risk measures, aggregation, rolling computations and
    plotting without raising. All returns, weights, composition, and derived measures
    are NaN.

    .. note::

        In backtesting workflows, when an optimization estimator is configured
        with `raise_on_failure=False`, a `FailedPortfolio` is returned on failed
        rebalancings. This lets the process complete without raising while preserving
        the full timeline for downstream analysis and diagnostics.

    Parameters
    ----------
    X : array-like of shape (n_observations, n_assets)
        Price returns of the assets.
        If `X` is a DataFrame, the columns will be considered as assets names
        and the indices will be considered as observations. Otherwise, we use
        `["x0", "x1", ..., "x(n_assets - 1)"]` as asset names and
        `[0, 1, ..., n_observations]` as observations.

    optimization_error : str, optional
        Stringified error message explaining why the optimization failed.
        Propagated from the optimization estimator when `raise_on_failure=False`.
        `None` means the reason is unknown or not provided.

    name : str, optional
        Name of the portfolio. The default (`None`) is to use the object id.

    tag : str, optional
        Tag given to the portfolio. Tags are used to manipulate groups of
        Portfolios from a `Population`.

    fallback_chain : list[tuple[str, str]] | None, optional
        Sequence describing the optimization fallback attempts. Each element is
        a pair `(estimator_repr, outcome)` where:

        * `estimator_repr` is the string representation of the primary
          estimator or a fallback (e.g. `"EqualWeighted()"`,
          `"previous_weights"`).
        * `outcome` is `"success"` if that step produced a valid solution,
          otherwise the stringified error message.

        For successful fits without any fallback, this is `None`. When
        fallbacks are provided and the primary fails, the chain starts with
        `(primary_repr, primary_error)` and is followed by one entry per
        fallback that was attempted, ending with the first `"success"` or the
        last error if all fail. This is set by the optimization estimator and
        propagated to the resulting portfolio objects (including
        `FailedPortfolio`).

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    risk_free_rate : float, default=0.0
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    annualized_factor : float, default=252.0
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    fitness_measures : list[measures], optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    compounded : bool, default=False
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    sample_weight : ndarray of shape (n_observations, ), optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    min_acceptable_return : float | None, optional
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    value_at_risk_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    entropic_risk_measure_theta : float, default=1.0
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    entropic_risk_measure_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    cvar_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    evar_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    drawdown_at_risk_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    cdar_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    edar_beta : float, default=0.95
        Accepted for API compatibility with `Portfolio` but not used by
        `FailedPortfolio`.

    Notes
    -----
    All performance, risk, and contribution measures are computed from NaN
    returns and NaN weights in a `FailedPortfolio`. As a result, these
    parameters do not affect the outcome: NaNs are carried over to metrics,
    contributions, plots, and rolling computations. This class exists solely to
    preserve API and type compatibility while signaling a failed optimization.
    """

    __slots__ = {
        # read-write
        "optimization_error",
    }

    def __init__(
        self,
        X: npt.ArrayLike,
        name: str | None = None,
        tag: str | None = None,
        optimization_error: str | None = None,
        fallback_chain: list[tuple[str, str]] | None = None,
        previous_weights: skt.MultiInput = None,
        transaction_costs: skt.MultiInput = None,
        management_fees: skt.MultiInput = None,
        risk_free_rate: float = 0,
        annualized_factor: float = 252,
        fitness_measures: list[skt.Measure] | None = None,
        compounded: bool = False,
        sample_weight: np.ndarray | None = None,
        min_acceptable_return: float | None = None,
        value_at_risk_beta: float = 0.95,
        entropic_risk_measure_theta: float = 1,
        entropic_risk_measure_beta: float = 0.95,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        drawdown_at_risk_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
    ):
        super().__init__(
            X=X,
            weights=None,
            previous_weights=previous_weights,
            transaction_costs=transaction_costs,
            management_fees=management_fees,
            risk_free_rate=risk_free_rate,
            name=name,
            tag=tag,
            annualized_factor=annualized_factor,
            fitness_measures=fitness_measures,
            compounded=compounded,
            sample_weight=sample_weight,
            min_acceptable_return=min_acceptable_return,
            value_at_risk_beta=value_at_risk_beta,
            entropic_risk_measure_theta=entropic_risk_measure_theta,
            entropic_risk_measure_beta=entropic_risk_measure_beta,
            cvar_beta=cvar_beta,
            evar_beta=evar_beta,
            drawdown_at_risk_beta=drawdown_at_risk_beta,
            cdar_beta=cdar_beta,
            edar_beta=edar_beta,
            fallback_chain=fallback_chain,
        )

        self.optimization_error = optimization_error



"""Hierarchical Equal Risk Contribution estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Weight constraints is a novel implementation, see docstring for more details.

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.cluster.hierarchy as sch
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

import skfolio.typing as skt
from skfolio.cluster import HierarchicalClustering
from skfolio.distance import BaseDistance, PearsonDistance
from skfolio.measures import ExtraRiskMeasure, RiskMeasure
from skfolio.optimization.cluster.hierarchical._base import (
    BaseHierarchicalOptimization,
)
from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.utils.stats import minimize_relative_weight_deviation
from skfolio.utils.tools import check_estimator


class HierarchicalEqualRiskContribution(BaseHierarchicalOptimization):
    r"""Hierarchical Equal Risk Contribution estimator.

    The Hierarchical Equal Risk Contribution is a portfolio optimization method
    developed by Thomas Raffinot [2]_.

    This algorithm uses a distance matrix to compute hierarchical clusters using the
    Hierarchical Tree Clustering algorithm. It then computes, for each cluster, the
    total cluster risk of an inverse-risk allocation.

    The final step is the top-down recursive division of the dendrogram, where the
    assets weights are updated using a naive risk parity within clusters.

    It differs from the Hierarchical Risk Parity by exploiting the dendrogram shape
    during the top-down recursive division instead of bisecting it.

    .. note ::

        The default linkage method is set to the Ward variance minimization algorithm,
        which is more stable and has better properties than the single-linkage
        method [4]_.

        Also, the initial paper does not provide an algorithm for handling weight
        constraints, and no standard solution currently exists.
        In contrast to HRP (Hierarchical Risk Parity), where weight constraints
        can be applied to the split factor at each bisection step, HERC
        (Hierarchical Equal Risk Contribution) cannot incorporate weight constraints
        during the intermediate steps of the allocation. Therefore, in HERC, the
        weight constraints must be enforced after the top-down allocation has been
        completed.
        In skfolio, we minimize the relative deviation of the final weights from
        the initial weights. This is formulated as a convex optimization problem:

        .. math::
            \begin{cases}
            \begin{aligned}
            &\min_{w} & & \Vert \frac{w - w_{init}}{w_{init}} \Vert_{2}^{2} \\
            &\text{s.t.} & & \sum_{i=1}^{N} w_{i} = 1 \\
            & & & w_{min} \leq w_i \leq w_{max}, \quad \forall i
            \end{aligned}
            \end{cases}

        The reason for minimizing the relative deviation (as opposed to the absolute
        deviation) is that we want to limit the impact on the risk contribution of
        each asset. Since HERC allocates inversely to risk, adjusting the weights
        based on relative deviation ensures that the assets' risk contributions
        remain proportionally consistent with the initial allocation.

    Parameters
    ----------
    risk_measure : RiskMeasure or ExtraRiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` or :class:`~skfolio.meta.ExtraRiskMeasure`
        of the optimization.
        Can be any of:

            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * VARIANCE
            * SEMI_VARIANCE
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO
            * VALUE_AT_RISK
            * DRAWDOWN_AT_RISK
            * ENTROPIC_RISK_MEASURE
            * FOURTH_CENTRAL_MOMENT
            * FOURTH_LOWER_PARTIAL_MOMENT

        The default is `RiskMeasure.VARIANCE`.

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.ReturnDistribution`
        containing the estimation of assets expected returns, covariance matrix and
        returns. The moments and returns estimations are used for the risk computation
        and the returns estimation are used by the distance matrix estimator.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    distance_estimator : BaseDistance, optional
        :ref:`Distance estimator <distance>`.
        The distance estimator is used to estimate the codependence and the distance
        matrix needed for the computation of the linkage matrix.
        The default (`None`) is to use :class:`~skfolio.distance.PearsonDistance`.

    hierarchical_clustering_estimator : HierarchicalClustering, optional
        :ref:`Hierarchical Clustering estimator <hierarchical_clustering>`.
        The hierarchical clustering estimator is used to compute the linkage matrix
        and the hierarchical clustering of the assets based on the distance matrix.
        The default (`None`) is to use
        :class:`~skfolio.cluster.HierarchicalClustering`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Minimum assets weights (weights lower bounds). The default is 0.0 (no short
        selling). Negative weights are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `0.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        minimum weight) and the input `X` of the `fit` methods must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default  minimum weight of `0.0`.

        Example:

           * `min_weights = 0.0` --> long only portfolio (default).
           * `min_weights = {"SX5E": 0.1, "SPX": 0.2}`
           * `min_weights = [0.1, 0.2]`

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=1.0
        Maximum assets weights (weights upper bounds). The default is 1.0 (each asset
        is below 100%). Weights above 1.0 are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `1.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        maximum weight) and the input `X` of the `fit` method must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default maximum weight of `1.0`.

        Example:

           * `max_weights = 1.0` --> each weight  must be below 100% (default).
           * `max_weights = 0.5` --> each weight must be below 50%.
           * `max_weights = {"SX5E": 0.8, "SPX": 0.9}`
           * `max_weights = [0.8, 0.9]`

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the
        optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the
        optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio total cost. If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the asset names in columns.
        The default (`None`) means no previous weights.
        Additionally, when `fallback="previous_weights"`, failures will fall back to
        these weights if provided.

    solver : str, default="CLARABEL"
        The solver used for the weights constraints optimization. The default is
        "CLARABEL" which is written in Rust and has better numerical stability and
        performance than ECOS and SCS.
        For more details about available solvers, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#choosing-a-solver

    solver_params : dict, optional
        Solver parameters. For example, `solver_params=dict(verbose=True)`.
        The default (`None`) is to use the CVXPY default.
        For more details about solver arguments, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options

    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        `transaction_costs`, `management_fees`, `previous_weights` and `risk_free_rate`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"` 
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary 
        estimator so that `fit` still returns the original instance. For traceability, 
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,)
        Weights of the assets.

    distance_estimator_ : BaseDistance
        Fitted `distance_estimator`.

    hierarchical_clustering_estimator_ : HierarchicalClustering
        Fitted `hierarchical_clustering_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has asset names that are all strings.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.

    References
    ----------
    .. [1]  "Hierarchical clustering-based asset allocation",
        The Journal of Portfolio Management,
        Thomas Raffinot  (2017).

    .. [2] "The hierarchical equal risk contribution portfolio",
        Thomas Raffinot (2018).

    .. [3] "Application of two-order difference to gap statistic".
        Yue, Wang & Wei (2009).

    .. [4] "A review of two decades of correlations, hierarchies, networks and
        clustering in financial markets",
        Gautier Marti, Frank Nielsen, Mikołaj Bińkowski, Philippe Donnat (2020).
    """

    def __init__(
        self,
        risk_measure: RiskMeasure | ExtraRiskMeasure = RiskMeasure.VARIANCE,
        prior_estimator: BasePrior | None = None,
        distance_estimator: BaseDistance | None = None,
        hierarchical_clustering_estimator: HierarchicalClustering | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        solver: str = "CLARABEL",
        solver_params: dict | None = None,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            risk_measure=risk_measure,
            prior_estimator=prior_estimator,
            distance_estimator=distance_estimator,
            hierarchical_clustering_estimator=hierarchical_clustering_estimator,
            min_weights=min_weights,
            max_weights=max_weights,
            transaction_costs=transaction_costs,
            management_fees=management_fees,
            previous_weights=previous_weights,
            portfolio_params=portfolio_params,
            fallback=fallback,
            raise_on_failure=raise_on_failure,
        )
        self.solver = solver
        self.solver_params = solver_params

    def fit(
        self, X: npt.ArrayLike, y: None = None, **fit_params
    ) -> HierarchicalEqualRiskContribution:
        """Fit the Hierarchical Equal Risk Contribution estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        **fit_params : dict
            Parameters to pass to the underlying estimators.
            Only available if `enable_metadata_routing=True`, which can be
            set by using ``sklearn.set_config(enable_metadata_routing=True)``.
            See :ref:`Metadata Routing User Guide <metadata_routing>` for
            more details.

        Returns
        -------
        self : HierarchicalEqualRiskContribution
            Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        # Validate
        if not isinstance(self.risk_measure, RiskMeasure | ExtraRiskMeasure):
            raise TypeError(
                "`risk_measure` must be of type `RiskMeasure` or `ExtraRiskMeasure`"
            )

        if self.risk_measure in [ExtraRiskMeasure.SKEW, ExtraRiskMeasure.KURTOSIS]:
            # Because Skew and Kurtosis can take negative values
            raise ValueError(
                f"risk_measure {self.risk_measure} currently not supported in HERC"
            )

        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.distance_estimator_ = check_estimator(
            self.distance_estimator,
            default=PearsonDistance(),
            check_type=BaseDistance,
        )
        self.hierarchical_clustering_estimator_ = check_estimator(
            self.hierarchical_clustering_estimator,
            default=HierarchicalClustering(),
            check_type=HierarchicalClustering,
        )

        # Fit the estimators
        self.prior_estimator_.fit(X, y, **routed_params.prior_estimator.fit)
        return_distribution = self.prior_estimator_.return_distribution_
        returns = return_distribution.returns

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            returns = pd.DataFrame(returns, columns=X.columns)

        # noinspection PyArgumentList
        self.distance_estimator_.fit(returns, y, **routed_params.distance_estimator.fit)
        distance = self.distance_estimator_.distance_

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            distance = pd.DataFrame(distance, columns=X.columns)

        # noinspection PyArgumentList
        self.hierarchical_clustering_estimator_.fit(
            X=distance, y=None, **routed_params.hierarchical_clustering_estimator.fit
        )

        n_clusters = self.hierarchical_clustering_estimator_.n_clusters_
        labels = self.hierarchical_clustering_estimator_.labels_
        linkage_matrix = self.hierarchical_clustering_estimator_.linkage_matrix_

        X = skv.validate_data(self, X)
        n_assets = X.shape[1]

        min_weights, max_weights = self._convert_weights_bounds(n_assets=n_assets)

        assets_risks = self._unitary_risks(return_distribution=return_distribution)
        weights = np.ones(n_assets)
        clusters_weights = np.ones(n_clusters)

        clusters = [np.argwhere(labels == i).flatten() for i in range(n_clusters)]
        clusters_sets = [set(cluster_ids) for cluster_ids in clusters]

        # Compute cluster total risk based on inverse-risk allocation
        cluster_risks = []
        for cluster_ids in clusters:
            inv_risk_w = np.zeros(n_assets)
            inv_risk_w[cluster_ids] = 1 / assets_risks[cluster_ids]
            inv_risk_w /= inv_risk_w.sum()
            cluster_risks.append(
                self._risk(weights=inv_risk_w, return_distribution=return_distribution)
            )
            weights[cluster_ids] = inv_risk_w[cluster_ids]
        cluster_risks = np.array(cluster_risks)

        # Compute the cluster weights using the dendrogram structure.
        # Recurse from the root until each of the defined cluster is reached and
        # update the weights using the naive risk parity.
        def _recurse(node):
            # Stop when the cluster is reached
            if set(node.pre_order()) in clusters_sets:
                return

            left_node = node.get_left()
            right_node = node.get_right()
            left_cluster_tree = set(left_node.pre_order())
            right_cluster_tree = set(right_node.pre_order())

            left_cluster = []
            right_cluster = []
            for i, cluster_ids in enumerate(clusters_sets):
                if cluster_ids.issubset(left_cluster_tree):
                    left_cluster.append(i)
                elif cluster_ids.issubset(right_cluster_tree):
                    right_cluster.append(i)

            if not left_cluster or not right_cluster:
                raise ValueError("Corrupted")

            left_cluster = np.array(left_cluster)
            right_cluster = np.array(right_cluster)

            left_risk = np.sum(cluster_risks[left_cluster])
            right_risk = np.sum(cluster_risks[right_cluster])

            alpha = 1 - left_risk / (left_risk + right_risk)

            clusters_weights[left_cluster] *= alpha
            clusters_weights[right_cluster] *= 1 - alpha

            _recurse(left_node)
            _recurse(right_node)

        root = sch.to_tree(linkage_matrix)
        _recurse(root)

        # Combine intra-cluster weights with inter-cluster weights
        for i, cluster_ids in enumerate(clusters):
            weights[cluster_ids] *= clusters_weights[i]

        # Apply weights constraints
        weights = minimize_relative_weight_deviation(
            weights=weights,
            min_weights=min_weights,
            max_weights=max_weights,
            solver=self.solver,
            solver_params=self.solver_params,
        )

        self.weights_ = weights

        return self



"""Base Bivariate Copula Estimator."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import sklearn.utils as sku
import sklearn.utils.validation as skv

from skfolio.distribution._base import BaseDistribution
from skfolio.distribution.copula._utils import (
    empirical_tail_concentration,
    plot_tail_concentration,
)

UNIFORM_MARGINAL_EPSILON = 1e-9
_RHO_BOUNDS = (-0.999, 0.999)


class BaseBivariateCopula(BaseDistribution, ABC):
    """Base class for Bivariate Copula Estimators.

    This abstract class defines the interface for bivariate copula models, including
    methods for fitting, sampling, scoring, and computing partial derivatives.

    Parameters
    ----------
    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.
    """

    # Used for AIC and BIC
    _n_params: int

    def __init__(self, random_state: int | None = None):
        super().__init__(random_state=random_state)

    def _validate_X(self, X: npt.ArrayLike, reset: bool) -> np.ndarray:
        """Validate the input data.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`.

        reset : bool, default=True
            Whether to reset the `n_features_in_` attribute.
            If False, the input will be checked for consistency with data
            provided when reset was last True.

        Returns
        -------
        validated_X: ndarray of shape (n_observations, 2)
            The validated data array.

        Raises
        ------
        ValueError
            If input data is invalid (e.g., not in `[0, 1]` or incorrect shape).
        """
        X = skv.validate_data(self, X, dtype=np.float64, reset=reset)
        if X.shape[1] != 2:
            raise ValueError("X must contains two columns for Bivariate Copula")
        if not np.all((X >= 0) & (X <= 1)):
            raise ValueError(
                "X must be in the interval `[0, 1]`, usually reprinting uniform "
                "distributions obtained from marginals CDF transformation"
            )

        # Handle potential numerical issues by ensuring X doesn't contain exact 0 or 1.
        X = np.clip(X, UNIFORM_MARGINAL_EPSILON, 1 - UNIFORM_MARGINAL_EPSILON)
        return X

    @property
    def n_params(self) -> int:
        """Number of model parameters."""
        return self._n_params

    @property
    @abstractmethod
    def lower_tail_dependence(self) -> float:
        """Theoretical lower tail dependence coefficient."""
        pass

    @property
    @abstractmethod
    def upper_tail_dependence(self) -> float:
        """Theoretical upper tail dependence coefficient."""
        pass

    @property
    @abstractmethod
    def fitted_repr(self) -> str:
        """String representation of the fitted copula."""
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None) -> "BaseBivariateCopula":
        """Fit the copula model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : BaseBivariateCopula
            Returns the instance itself.
        """
        pass

    @abstractmethod
    def cdf(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the CDF of the bivariate copula.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        Returns
        -------
        cdf : ndarray of shape (n_observations,)
            CDF values for each observation in X.
        """
        pass

    @abstractmethod
    def partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the h-function (partial derivative) for the bivariate copula
        with respect to a specified margin.

        The h-function with respect to the second margin represents the conditional
        distribution function of :math:`u` given :math:`v`:

        .. math::
            h(u \mid v) = \frac{\partial C(u,v)}{\partial v}

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        first_margin : bool, default=False
            If True, compute the partial derivative with respect to the first
            margin `u`; otherwise, compute the partial derivative with respect to the
            second margin `v`.

        Returns
        -------
        p : ndarray of shape (n_observations,)
            h-function values :math:`h(u \mid v) \;=\; p` for each observation in X.
        """
        pass

    @abstractmethod
    def inverse_partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the inverse of the bivariate copula's partial derivative, commonly
        known as the inverse h-function [1]_.

        Let :math:`C(u, v)` be a bivariate copula. The h-function with respect to the
        second margin is defined by

        .. math::
            h(u \mid v) \;=\; \frac{\partial\,C(u, v)}{\partial\,v},

        which is the conditional distribution of :math:`U` given :math:`V = v`.
        The **inverse h-function**, denoted :math:`h^{-1}(p \mid v)`, is the unique
        value :math:`u \in [0,1]` such that

        .. math::
            h(u \mid v) \;=\; p,
            \quad \text{where } p \in [0,1].

        In practical terms, given :math:`(p, v)` in :math:`[0, 1]^2`,
        :math:`h^{-1}(p \mid v)` solves for the :math:`u` satisfying
        :math:`p = \partial C(u, v)/\partial v`.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(p, v)`, each in the interval `[0, 1]`.
            - The first column `p` corresponds to the value of the h-function.
            - The second column `v` is the conditioning variable.

        first_margin : bool, default=False
            If True, compute the inverse partial derivative with respect to the first
            margin `u`; otherwise, compute the inverse partial derivative with respect
            to the second margin `v`.

        Returns
        -------
        u : ndarray of shape (n_observations,)
            A 1D-array of length `n_observations`, where each element is the computed
            :math:`u = h^{-1}(p \mid v)` for the corresponding pair in `X`.

        References
        ----------
        .. [1] "Multivariate Models and Dependence Concepts", Joe, H. (1997)
        .. [2] "An Introduction to Copulas", Nelsen, R. B. (2006)
        """
        pass

    @abstractmethod
    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the log-likelihood of each sample (log-pdf) under the model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            The log-likelihood of each sample under the fitted copula.
        """
        pass

    def sample(self, n_samples: int = 1):
        """Generate random samples from the bivariate copula using the inverse
        Rosenblatt transform.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        Returns
        -------
        X : array-like of shape (n_samples, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` are uniform marginals in the
            interval `[0, 1]`.
        """
        skv.check_is_fitted(self)
        rng = sku.check_random_state(self.random_state)

        # Generate independent Uniform(0, 1) samples
        X = rng.random(size=(n_samples, 2))

        # Apply the inverse Rosenblatt transform on the first variable.
        X[:, 1] = self.inverse_partial_derivative(X, first_margin=True)
        return X

    def tail_concentration(self, quantiles: np.ndarray) -> np.ndarray:
        """
        Compute the tail concentration function for a set of quantiles.

        The tail concentration function is defined as follows:
         - For quantiles q ≤ 0.5:
             C(q) = P(U ≤ q, V ≤ q) / q

         - For quantiles q > 0.5:
             C(q) = (1 - 2q + P(U ≤ q, V ≤ q)) / (1 - q)

        where U and V are the pseudo-observations of the first and second variables,
        respectively. This function returns the concentration values for each q
        provided.

        Parameters
        ----------
        quantiles : ndarray of shape (n_quantiles,)
           A 1D array of quantile levels (values between 0 and 1) at which to compute
           the tail concentration.

        Returns
        -------
        concentration : ndarray of shape (n_quantiles,)
           The computed tail concentration values corresponding to each quantile.

        References
        ----------
        .. [1] "Quantitative Risk Management: Concepts, Techniques, and Tools",
            McNeil, Frey, Embrechts (2005)

        Raises
        ------
        ValueError
           If any value in `quantiles` is not in the interval [0, 1].
        """
        quantiles = np.asarray(quantiles)
        if not np.all((quantiles >= 0) & (quantiles <= 1)):
            raise ValueError("quantiles must be between 0.0 and 1.0.")
        X = np.stack((quantiles, quantiles)).T
        cdf = self.cdf(X)
        concentration = np.where(
            quantiles <= 0.5,
            cdf / quantiles,
            (1.0 - 2 * quantiles + cdf) / (1.0 - quantiles),
        )
        return concentration

    def plot_tail_concentration(
        self, X: npt.ArrayLike | None = None, title: str | None = None
    ) -> go.Figure:
        """
        Plot the tail concentration function.

        This method computes the tail concentration function at 100 evenly spaced
        quantile levels between 0.005 and 0.995.
        The plot displays the concentration values on the y-axis and the quantile levels
        on the x-axis.

        The tail concentration is defined as:
          - Lower tail: λ_L(q) = P(U₂ ≤ q | U₁ ≤ q)
          - Upper tail: λ_U(q) = P(U₂ ≥ q | U₁ ≥ q)

        where U₁ and U₂ are the pseudo-observations of the first and second variables,
        respectively.

        Parameters
        ----------
        X : array-like of shape (n_samples, 2), optional
            If provided, it is used to plot the empirical tail concentration for
            comparison versus the model tail concentration.

        title : str, optional
            The title for the plot. If not provided, a default title based on the fitted
            copula's representation is used.

        Returns
        -------
        fig : go.Figure
            A Plotly figure object containing the tail concentration curve.

        References
        ----------
        .. [1] "Quantitative Risk Management: Concepts, Techniques, and Tools",
            McNeil, Frey, Embrechts (2005)
        """
        if title is None:
            title = f"Tail Concentration of Bivariate {self.__class__.__name__}"
            if X is not None:
                title += " vs Empirical"

        quantiles = np.linspace(5e-3, 1.0 - 5e-3, num=100)
        concentration = self.tail_concentration(quantiles)

        tail_concentration_dict = {self.__class__.__name__: concentration}
        if X is not None:
            tail_concentration_dict["Empirical"] = empirical_tail_concentration(
                X, quantiles=quantiles
            )

        fig = plot_tail_concentration(
            tail_concentration_dict=tail_concentration_dict,
            quantiles=quantiles,
            title=title,
            smoothing=1.3,
        )
        return fig

    def plot_pdf_2d(self, title: str | None = None) -> go.Figure:
        """
        Plot a 2D contour of the estimated probability density function (PDF).

        This method generates a grid over [0, 1]^2, computes the PDF, and displays a
        contour plot of the PDF.
        Contour levels are limited to the 97th quantile to avoid extreme densities.

        Parameters
        ----------
        title : str, optional
           The title for the plot. If not provided, a default title based on the fitted
           copula's representation is used.

        Returns
        -------
        fig : go.Figure
           A Plotly figure object containing the 2D contour plot of the PDF.
        """
        skv.check_is_fitted(self)

        if title is None:
            title = f"PDF of the Bivariate {self.__class__.__name__}"

        u = np.linspace(0.01, 0.99, 100)
        U, V = np.meshgrid(u, u)
        grid_points = np.column_stack((U.ravel(), V.ravel()))
        pdfs = np.exp(self.score_samples(grid_points)).reshape(U.shape)
        # After the 97th quantile, the pdf gets too dense, and it dilutes the plot.
        end = round(np.quantile(pdfs, 0.97), 1)
        fig = go.Figure(
            data=go.Contour(
                x=u,
                y=u,
                z=pdfs,
                colorscale="Magma",
                contours=dict(start=0, end=end, size=0.2),
                line=dict(width=0),
                colorbar=dict(title="PDF"),
            )
        )
        fig.update_layout(
            title=title,
            xaxis_title="u",
            yaxis_title="v",
        )
        return fig

    def plot_pdf_3d(self, title: str | None = None) -> go.Figure:
        """
        Plot a 3D surface of the estimated probability density function (PDF).

        This method generates a grid over [0, 1]^2, computes the PDF, and displays a
        3D surface plot of the PDF using Plotly.

        Parameters
        ----------
        title : str, optional
           The title for the plot. If not provided, a default title based on the fitted
           copula's representation is used.

        Returns
        -------
        fig : go.Figure
           A Plotly figure object containing a 3D surface plot of the PDF.
        """
        skv.check_is_fitted(self)

        if title is None:
            title = f"PDF of the Bivariate {self.__class__.__name__}"

        u = np.linspace(0.03, 0.97, 100)
        U, V = np.meshgrid(u, u)
        grid_points = np.column_stack((U.ravel(), V.ravel()))
        pdfs = np.exp(self.score_samples(grid_points)).reshape(U.shape)
        fig = go.Figure(data=[go.Surface(x=U, y=V, z=pdfs, colorscale="Magma")])
        fig.update_layout(
            title=title, scene=dict(xaxis_title="u", yaxis_title="v", zaxis_title="PDF")
        )
        return fig



"""Normal Inverse Gaussian Estimator."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy.typing as npt
import scipy.stats as st

from skfolio.distribution.univariate._base import BaseUnivariateDist


class NormalInverseGaussian(BaseUnivariateDist):
    r"""Normal Inverse Gaussian Distribution Estimation.

    This estimator fits a univariate Normal Inverse Gaussian (NIG) distribution
    to the input data.

    The probability density function is:

    .. math::

        f(x, a, b) = \frac{a \, K_1(a \sqrt{1 + x^2})}{\pi \sqrt{1 + x^2}} \,
                     \exp(\sqrt{a^2 - b^2} + b x)

    where :math:`x` is a real number, the parameter :math:`a` is the tail
    heaviness and :math:`b` is the asymmetry parameter satisfying :math:`a > 0`
    and :math:`|b| <= a`. :math:`K_1` is the modified Bessel function of second kind
    (`scipy.special.k1`).

    The probability density above is defined in the "standardized" form. To shift
    and/or scale the distribution use the loc and scale parameters. Specifically,
    `pdf(x, a, b, loc, scale)` is equivalent to `pdf(y, a, b) / scale` with
    `y = (x - loc) / scale`.

    For more information, you can refer to the `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norminvgauss.html#scipy.stats.norminvgauss>`_

    Parameters
    ----------
    loc : float, optional
        If provided, the location parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    scale : float, optional
        If provided, the scale parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    a_ : float
        The fitted shape parameter a of the NIG distribution.

    b_ : float
        The fitted shape parameter b of the NIG distribution.

    loc_ : float
        The fitted location parameter.

    scale_ : float
        The fitted scale parameter.

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_index
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution.univariate import NormalInverseGaussian
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_index()
    >>> X = prices_to_returns(prices)
    >>>
    >>> # Initialize the estimator.
    >>> model = NormalInverseGaussian()
    >>>
    >>> # Fit the model to the data.
    >>> model.fit(X)
    >>>
    >>> # Display the fitted parameters.
    >>> print(model.fitted_repr)
    NormalInverseGaussian(0.422, -0.0321, 0.000913, 0.00739)
    >>>
    >>> # Compute the log-likelihood, total log-likelihood, CDF, PPF, AIC, and BIC
    >>> log_likelihood = model.score_samples(X)
    >>> score = model.score(X)
    >>> cdf = model.cdf(X)
    >>> ppf = model.ppf(X)
    >>> aic = model.aic(X)
    >>> bic = model.bic(X)
    >>>
    >>> # Generate 5 new samples from the fitted distribution.
    >>> samples = model.sample(n_samples=5)
    >>>
    >>> # Plot the estimated probability density function (PDF).
    >>> fig = model.plot_pdf()
    >>> fig.show()
    """

    a_: float
    b_: float
    loc_: float
    scale_: float
    _scipy_model = st.norminvgauss

    def __init__(
        self,
        loc: float | None = None,
        scale: float | None = None,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.loc = loc
        self.scale = scale

    @property
    def _scipy_params(self) -> dict[str, float]:
        """Dictionary of parameters to pass to the underlying SciPy distribution."""
        return {"a": self.a_, "b": self.b_, "loc": self.loc_, "scale": self.scale_}

    def fit(self, X: npt.ArrayLike, y=None) -> "NormalInverseGaussian":
        """Fit the univariate Normal Inverse Gaussian distribution model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            The input data. X must contain a single column.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : NormalInverseGaussian
            Returns the instance itself.
        """
        X = self._validate_X(X, reset=True)

        if self.loc is not None and self.scale is not None:
            raise ValueError("Either loc or scale must be None to be fitted")

        fixed_params = {}
        if self.loc is not None:
            fixed_params["floc"] = self.loc
        if self.scale is not None:
            fixed_params["fscale"] = self.scale

        self.a_, self.b_, self.loc_, self.scale_ = self._scipy_model.fit(
            X, **fixed_params
        )
        return self



"""Preprocessing module to transform X to returns."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from typing import Literal

import numpy as np
import pandas as pd


def prices_to_returns(
    X: pd.DataFrame,
    y: pd.DataFrame | None = None,
    log_returns: bool = False,
    nan_threshold: float = 1,
    join: Literal["left", "right", "inner", "outer", "cross"] = "outer",
    drop_inceptions_nan: bool = True,
    fill_nan: bool = True,
) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
    r"""Transform a DataFrame of prices to linear or logarithmic returns.

    Linear returns (also called simple returns) are defined as:
        .. math:: \frac{S_{t}}{S_{t-1}} - 1

    Logarithmic returns (also called continuously compounded return) are defined as:
        .. math:: ln\Biggl(\frac{S_{t}}{S_{t-1}}\Biggr)

    With :math:`S_{t}` the asset price at time :math:`t`.

    .. warning::

        The linear returns aggregate across securities, meaning that the linear return
        of the portfolio is the weighted average of the linear returns of the
        securities. For this reason, **portfolio optimization should be performed
        using linear returns** [1]_.

        On the other hand, the logarithmic returns aggregate across time, meaning that
        the total logarithmic return over K time periods is the sum of all K
        single-period logarithmic returns.

    .. seealso::

        :ref:`data preparation <data_preparation>`

    Parameters
    ----------
    X : DataFrame
        The DataFrame of assets prices.

    y : DataFrame, optional
        The DataFrame of target or factors prices.
        If provided, it is joined with the DataFrame of prices to ensure identical
        observations.

    log_returns : bool, default=True
        If this is set to True, logarithmic returns are used instead of simple returns.

    join : str, default="outer"
        The join method between `X` and `y` when `y` is provided.

    nan_threshold : float, default=1.0
        Drop observations (rows) that have a percentage of missing assets prices above
        this threshold. The default (`1.0`) is to keep all the observations.

    drop_inceptions_nan : bool, default=True
        If set to True, observations at the beginning are dropped if any of
        the asset values are missing, otherwise we keep the NaNs. This is useful when
        you work with a large universe of assets with different inception dates coupled
        with a pre-selection Transformer.

    fill_nan : bool, default=True
        If set to True, missing prices (NaNs) are forward filled using the previous
        price. Otherwise, NaNs are kept.

    Returns
    -------
    X : DataFrame
        The DataFrame of price returns of the input `X`.

    y : DataFrame, optional
        The DataFrame of price returns of the input `y` when provided.

    References
    ----------
    .. [1]  "Linear vs. Compounded Returns - Common Pitfalls in Portfolio Management".
        GARP Risk Professional.
        Attilio Meucci (2010).
    """
    if not isinstance(X, pd.DataFrame):
        raise TypeError("`X` must be a DataFrame")

    if y is None:
        df = X.copy()
    else:
        if not isinstance(y, pd.DataFrame):
            raise TypeError("`y` must be a DataFrame")
        df = X.join(y, how=join)

    _, n_assets = X.shape

    # Remove observations with missing X above threshold
    if nan_threshold is not None:
        nan_threshold = float(nan_threshold)
        if not 0 < nan_threshold <= 1:
            raise ValueError("`nan_threshold` must be between 0 and 1")
        count_nan = df.isna().sum(axis=1)
        to_drop = count_nan[count_nan > n_assets * nan_threshold].index
        if len(to_drop) > 0:
            df.drop(to_drop, axis=0, inplace=True)

    # Forward fill missing values
    if fill_nan:
        df.ffill(inplace=True)
    # Drop rows according to drop_inceptions_nan
    # noinspection PyTypeChecker
    df.dropna(how="any" if drop_inceptions_nan else "all", inplace=True)
    # Drop column if all its values are missing
    df.dropna(axis=1, how="all", inplace=True)

    # returns
    all_returns = df.pct_change(fill_method=None).iloc[1:]
    if log_returns:
        all_returns = np.log1p(all_returns)

    if y is None:
        return all_returns

    returns = all_returns[[x for x in X.columns if x in df.columns]]
    factor_returns = all_returns[[x for x in y.columns if x in df.columns]]
    return returns, factor_returns



"""Pre-selection DropCorrelated module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
import sklearn.feature_selection as skf
import sklearn.utils.validation as skv


class DropCorrelated(skf.SelectorMixin, skb.BaseEstimator):
    """Transformer for dropping highly correlated assets.

    Simply removing all correlation pairs above the threshold will remove more assets
    than necessary and a naive sequential removal is suboptimal and depends on the
    initial assets ordering.

    Let's suppose X,Y,Z are three random variables with corr(X,Y) and corr(X,Z) above
    the threshold and corr(Y,Z) below.
    The first approach would remove X,Y,Z and the second approach would remove either
    Y and Z or X depending on the initial ordering.

    To avoid these shortcomings, we implement the below algorithm:

        * Step 1: select all correlation pairs above the threshold.
        * Step 2: sort all the selected correlation pairs from highest to lowest.
        * Step 3: for each pair, if none of the two assets has been removed, keep the
          asset with the lowest average correlation against the other assets.

    Parameters
    ----------
    threshold : float, default=0.95
        Correlation threshold. The default value is `0.95`.

    absolute : bool, default=False
        If this is set to True, we take the absolute value of the correlation. This has
        for effect to also include negatively correlated assets.

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
        Boolean array indicating which assets are remaining.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    to_keep_: np.ndarray

    def __init__(self, threshold: float = 0.95, absolute: bool = False):
        self.threshold = threshold
        self.absolute = absolute

    def fit(self, X: npt.ArrayLike, y=None):
        """Run the correlation transformer and get the appropriate assets.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : DropCorrelated
            Fitted estimator.
        """
        X = skv.validate_data(self, X)
        if not -1 <= self.threshold <= 1:
            raise ValueError("`threshold` must be between -1 and 1")

        n_assets = X.shape[1]
        corr = np.corrcoef(X.T)
        mean_corr = corr.mean(axis=0)

        triu_idx = np.triu_indices(n_assets, 1)

        # select all correlation pairs above the threshold
        selected_idx = np.argwhere(corr[triu_idx] > self.threshold).flatten()

        # sort all the selected correlation pairs from highest to lowest
        selected_idx = selected_idx[np.argsort(-corr[triu_idx][selected_idx])]

        # for each pair, if none of the two assets has been removed, keep the asset with
        # the lowest average correlation with other assets
        to_remove = set()
        for idx in selected_idx:
            i, j = triu_idx[0][idx], triu_idx[1][idx]
            if i not in to_remove and j not in to_remove:
                if mean_corr[i] > mean_corr[j]:
                    to_remove.add(i)
                else:
                    to_remove.add(j)
        self.to_keep_ = ~np.isin(np.arange(n_assets), list(to_remove))
        return self

    def _get_support_mask(self):
        skv.check_is_fitted(self)
        return self.to_keep_



"""LedoitWolf Covariance Estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import numpy.typing as npt
import sklearn.covariance as skc

from skfolio.moments.covariance._base import BaseCovariance


class LedoitWolf(BaseCovariance, skc.LedoitWolf):
    """LedoitWolf Covariance Estimator.

    Ledoit-Wolf is a particular form of shrinkage, where the shrinkage
    coefficient is computed using O. Ledoit and M. Wolf's formula as
    described in [1]_.

    Read more in `scikit-learn
    <https://scikit-learn.org/stable/modules/generated/sklearn.covariance.ShrunkCovariance.html>`_.

    Parameters
    ----------
    store_precision : bool, default=True
        Specify if the estimated precision is stored.

    assume_centered : bool, default=False
        If True, data will not be centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False (default), data will be centered before computation.

    block_size : int, default=1000
        Size of blocks into which the covariance matrix will be split
        during its Ledoit-Wolf estimation. This is purely a memory
        optimization and does not affect results.

    nearest : bool, default=True
        If this is set to True, the covariance is replaced by the nearest covariance
        matrix that is positive definite and with a Cholesky decomposition than can be
        computed. The variance is left unchanged.
        A covariance matrix that is not positive definite often occurs in high
        dimensional problems. It can be due to multicollinearity, floating-point
        inaccuracies, or when the number of observations is smaller than the number of
        assets. For more details, see :func:`~skfolio.utils.stats.cov_nearest`.
        The default is `True`.

    higham : bool, default=False
        If this is set to True, the Higham & Nick (2002) algorithm is used to find the
        nearest PD covariance, otherwise the eigenvalues are clipped to a threshold
        above zeros (1e-13). The default is `False` and use the clipping method as the
        Higham & Nick algorithm can be slow for large datasets.

    higham_max_iteration : int, default=100
        Maximum number of iteration of the Higham & Nick (2002) algorithm.
        The default value is `100`.

    Attributes
    ----------
    covariance_ : ndarray of shape (n_assets, n_assets)
        Estimated covariance.

    location_ : ndarray of shape (n_assets,)
        Estimated location, i.e. the estimated mean.

    precision_ : ndarray of shape (n_assets, n_assets)
        Estimated pseudo inverse matrix.
        (stored only if store_precision is True)

    shrinkage_ : float
        Coefficient in the convex combination used for the computation
        of the shrunk estimate. Range is [0, 1].

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Notes
    -----
    The regularised covariance is:

    (1 - shrinkage) * cov + shrinkage * mu * np.identity(n_features)

    where mu = trace(cov) / n_features
    and shrinkage is given by the Ledoit and Wolf formula (see References)

    References
    ----------
    .. [1]  "A Well-Conditioned Estimator for Large-Dimensional Covariance Matrices".
        Ledoit and Wolf, Journal of Multivariate Analysis, Volume 88, Issue 2.
        February 2004, pages 365-41.
    """

    def __init__(
        self,
        store_precision=True,
        assume_centered=False,
        block_size=1000,
        nearest: bool = True,
        higham: bool = False,
        higham_max_iteration: int = 100,
    ):
        super().__init__(
            nearest=nearest,
            higham=higham,
            higham_max_iteration=higham_max_iteration,
        )
        skc.LedoitWolf.__init__(
            self,
            store_precision=store_precision,
            assume_centered=assume_centered,
            block_size=block_size,
        )

    def fit(self, X: npt.ArrayLike, y=None) -> "LedoitWolf":
        """Fit the Ledoit-Wolf shrunk covariance model to X.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
         Price returns of the assets.

        y : Ignored
          Not used, present for API consistency by convention.

        Returns
        -------
        self : LedoitWolf
          Fitted estimator.
        """
        skc.LedoitWolf.fit(self, X)
        self._set_covariance(self.covariance_)
        return self



"""pre-selection SelectComplete module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
import sklearn.feature_selection as skf
import sklearn.utils.validation as skv


class SelectComplete(skf.SelectorMixin, skb.BaseEstimator):
    """
    Transformer to select assets with complete data across the entire observation
    period.

    This transformer removes assets (columns) that have missing values (NaNs) at the
    beginning or end of the period.

    This transformer is especially useful for financial datasets where assets
    (e.g., stocks, bonds) may have data gaps due to late inception (assets that started
    trading later), early expiry or default (assets that stopped trading before the
    end of the period).

    If missing values are not at the beginning or end but occur between non-missing
    values, the asset is not removed unless `drop_assets_with_internal_nan` is set to
    `True`.

    Parameters
    ----------
    drop_assets_with_internal_nan : bool, default=False
        If set to True, assets with missing values (NaNs) that appear between
        non-missing values (i.e., internal NaNs) will also be removed. By default,
        only assets with leading or trailing NaNs are removed.

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
       Boolean array indicating which assets are remaining.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from skfolio.pre_selection import SelectComplete
    >>> X = pd.DataFrame({
    ...     'asset1': [np.nan, np.nan, 2, 3, 4],    # Starts late (inception)
    ...     'asset2': [1, 2, 3, 4, 5],         # Complete data
    ...     'asset3': [1, 2, 3, np.nan, 5], # Missing values within data
    ...     'asset4': [1, 2, 3, 4, np.nan]      # Ends early (expiration)
    ... })
    >>> selector = SelectComplete()
    >>> selector.fit_transform(X)
     array([[ 1.,  1.],
            [ 2.,  2.],
            [ 3.,  3.],
            [ 4., nan],
            [ 5.,  5.]])
    >>> selector = SelectComplete(drop_assets_with_internal_nan=True)
    >>> selector.fit_transform(X)
     array([[1.],
           [2.],
           [3.],
           [4.],
           [5.]])
    """

    to_keep_: np.ndarray

    def __init__(self, drop_assets_with_internal_nan: bool = False):
        self.drop_assets_with_internal_nan = drop_assets_with_internal_nan

    def fit(self, X: npt.ArrayLike, y=None) -> "SelectComplete":
        """Run the SelectComplete transformer and get the appropriate assets.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : SelectComplete
            Fitted estimator.
        """
        # Validate by allowing NaNs
        X = skv.validate_data(self, X, ensure_all_finite="allow-nan")

        if self.drop_assets_with_internal_nan:
            # Identify columns with any NaNs
            self.to_keep_ = ~np.isnan(X).any(axis=0)
        else:
            # Identify columns with no leading or trailing NaNs
            self.to_keep_ = ~np.isnan(X[0, :]) & ~np.isnan(X[-1, :])

        return self

    def _get_support_mask(self) -> np.ndarray:
        skv.check_is_fitted(self)
        return self.to_keep_

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        return tags



"""Base Multivariate Distribution Estimator."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import sklearn.utils as sku

from skfolio.distribution._base import BaseDistribution


class BaseMultivariateDist(BaseDistribution, ABC):
    """Base class for Multivariate Distribution Estimators.

    This abstract class defines the interface for multivariate distribution models.

    Parameters
    ----------
    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.
    """

    # Used for AIC and BIC
    _n_params: int

    def __init__(self, random_state: int | None = None):
        super().__init__(random_state=random_state)

    @property
    @abstractmethod
    def n_params(self) -> int:
        """Number of model parameters."""
        pass

    @property
    @abstractmethod
    def fitted_repr(self) -> str:
        """String representation of the fitted copula."""
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None) -> "BaseMultivariateDist":
        """Fit the multivariate distribution model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : BaseMultivariateDist
            Returns the instance itself.
        """
        pass

    @abstractmethod
    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the log-likelihood of each sample (log-pdf) under the distribution
        model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            The log-likelihood of each sample under the fitted distribution model.
        """
        pass

    @abstractmethod
    def sample(
        self,
        n_samples: int = 1,
        conditioning: dict[int | str : float | tuple[float, float] | npt.ArrayLike]
        | None = None,
    ) -> np.ndarray:
        """Generate random samples from the distribution model.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        conditioning : dict[int | str, float | tuple[float, float] | array-like], optional
            A dictionary specifying conditioning information for one or more assets.
            The dictionary keys are asset indices or names, and the values define how
            the samples are conditioned for that asset. Three types of conditioning
            values are supported:

            1. **Fixed value (float):**
               If a float is provided, all samples are generated under the condition
               that the asset takes exactly that value.

            2. **Bounds (tuple of two floats):**
               If a tuple `(min_value, max_value)` is provided, samples are generated
               under the condition that the asset's value falls within the specified
               bounds. Use `-np.Inf` for no lower bound or `np.Inf` for no upper bound.

            3. **Array-like (1D array):**
               If an array-like of length `n_samples` is provided, each sample is
               conditioned on the corresponding value in the array for that asset.

        Returns
        -------
        X : array-like of shape (n_samples, n_assets)
            A two-dimensional array where each row is a multivariate observation sampled
            from the fitted distribution model.
        """
        pass

    def plot_scatter_matrix(
        self,
        X: npt.ArrayLike | None = None,
        conditioning: dict[int | str : float | tuple[float, float] | npt.ArrayLike]
        | None = None,
        n_samples: int = 1000,
        title: str = "Scatter Matrix",
    ) -> go.Figure:
        """
        Plot the vine copula scatter matrix by generating samples from the fitted
        distribution model and comparing it versus the empirical distribution of `X` if
        provided.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_assets), optional
            If provided, it is used to plot the empirical scatter matrix for
            comparison versus the vine copula scatter matrix.

        conditioning : dict[int | str, float | tuple[float, float] | array-like], optional
            A dictionary specifying conditioning information for one or more assets.
            The dictionary keys are asset indices or names, and the values define how
            the samples are conditioned for that asset. Three types of conditioning
            values are supported:

            1. **Fixed value (float):**
               If a float is provided, all samples are generated under the condition
               that the asset takes exactly that value.

            2. **Bounds (tuple of two floats):**
               If a tuple `(min_value, max_value)` is provided, samples are generated
               under the condition that the asset's value falls within the specified
               bounds. Use `-np.Inf` for no lower bound or `np.Inf` for no upper bound.

            3. **Array-like (1D array):**
               If an array-like of length `n_samples` is provided, each sample is
               conditioned on the corresponding value in the array for that asset.

        n_samples : int, default=1000
            Number of samples used to control the density and readability of the plot.
            If `X` is provided and contains more than `n_samples` rows, a random
            subsample of size `n_samples` is selected. Conversely, if `X` has fewer
            rows than `n_samples`, the value is adjusted to match the number of rows in
            `X` to ensure balanced visualization.

        title : str, default="Scatter Matrix"
            The title for the plot.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            A figure object containing the scatter matrix.
        """
        traces = []
        n_assets = self.n_features_in_
        if X is not None:
            X = np.asarray(X)
            if X.ndim != 2:
                raise ValueError("X should be an 2D array")
            if X.shape[1] != n_assets:
                raise ValueError(f"X should have {n_assets} columns")
            if X.shape[0] > n_samples:
                # We subsample for improved graph readability
                rng = sku.check_random_state(self.random_state)
                indices = rng.choice(
                    np.arange(X.shape[0]), size=n_samples, replace=False
                )
                X = X[indices, :]
            else:
                # We want same proportion as X to have a balanced graph
                n_samples = X.shape[0]
            traces.append(
                go.Splom(
                    dimensions=[
                        {"label": self.feature_names_in_[i], "values": X[:, i]}
                        for i in range(n_assets)
                    ],
                    showupperhalf=False,
                    diagonal_visible=False,
                    marker=dict(
                        size=5,
                        color="rgb(85,168,104)",
                        line=dict(width=0.2, color="white"),
                        opacity=0.6,
                    ),
                    name="Historical",
                    showlegend=True,
                )
            )

        sample = self.sample(n_samples=n_samples, conditioning=conditioning)

        traces.append(
            go.Splom(
                dimensions=[
                    {"label": self.feature_names_in_[i], "values": sample[:, i]}
                    for i in range(n_assets)
                ],
                showupperhalf=False,
                diagonal_visible=False,
                marker=dict(
                    size=5,
                    color="rgb(221,132,82)",
                    line=dict(width=0.2, color="white"),
                    opacity=0.6,
                ),
                name="Generated",
                showlegend=True,
            )
        )

        if conditioning is not None:
            # Improve readability
            traces = traces[::-1]

        fig = go.Figure(data=traces)
        fig.update_layout(title=title)
        return fig



"""Pre Selection module."""

from skfolio.pre_selection._drop_correlated import DropCorrelated
from skfolio.pre_selection._drop_zero_variance import DropZeroVariance
from skfolio.pre_selection._select_complete import SelectComplete
from skfolio.pre_selection._select_k_extremes import SelectKExtremes
from skfolio.pre_selection._select_non_dominated import SelectNonDominated
from skfolio.pre_selection._select_non_expiring import SelectNonExpiring

__all__ = [
    "DropCorrelated",
    "DropZeroVariance",
    "SelectComplete",
    "SelectKExtremes",
    "SelectNonDominated",
    "SelectNonExpiring",
]



"""Ensemble Optimization module."""

from skfolio.optimization.ensemble._stacking import StackingOptimization
from skfolio.utils.composition import BaseComposition

__all__ = ["BaseComposition", "StackingOptimization"]



"""Univariate Distribution Selection."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import sklearn as sk

from skfolio.distribution._base import SelectionCriterion
from skfolio.distribution.univariate._base import BaseUnivariateDist
from skfolio.distribution.univariate._gaussian import Gaussian
from skfolio.distribution.univariate._johnson_su import JohnsonSU
from skfolio.distribution.univariate._student_t import StudentT


def select_univariate_dist(
    X: npt.ArrayLike,
    distribution_candidates: list[BaseUnivariateDist] | None = None,
    selection_criterion: SelectionCriterion = SelectionCriterion.AIC,
) -> BaseUnivariateDist:
    """Select the optimal univariate distribution estimator based on an information
    criterion.

    For each candidate distribution, the function fits the distribution to X and then
    computes either the Akaike Information Criterion (AIC) or the Bayesian Information
    Criterion (BIC). The candidate with the lowest criterion value is returned.

    Parameters
    ----------
    X : array-like of shape (n_observations, 1)
        The input data used to fit each candidate distribution.

    distribution_candidates : list of BaseUnivariateDist
        A list of candidate distribution estimators. Each candidate must be an instance
        of a class that inherits from `BaseUnivariateDist`.
        If None, defaults to `[Gaussian(), StudentT(), JohnsonSU()]`.

    selection_criterion : SelectionCriterion, default=SelectionCriterion.AIC
        The criterion used for model selection. Possible values are:
            - SelectionCriterion.AIC : Akaike Information Criterion
            - SelectionCriterion.BIC : Bayesian Information Criterion

    Returns
    -------
    BaseUnivariateDist
        The fitted candidate estimator that minimizes the selected information
        criterion.

    Raises
    ------
    ValueError
        If X does not have exactly one column or if any candidate in the list does not
        inherit from BaseUnivariateDist.
    """
    if distribution_candidates is None:
        distribution_candidates = [
            Gaussian(),
            StudentT(),
            JohnsonSU(),
        ]

    X = np.asarray(X)
    if X.ndim != 2 or X.shape[1] != 1:
        raise ValueError("X must contains one column for Univariate Distribution")

    results = {}
    for dist in distribution_candidates:
        if not isinstance(dist, BaseUnivariateDist):
            raise ValueError("Each candidate must inherit from `BaseUnivariateDist`")
        dist = sk.clone(dist)
        dist.fit(X)

        match selection_criterion:
            case selection_criterion.AIC:
                results[dist] = dist.aic(X)
            case selection_criterion.BIC:
                results[dist] = dist.bic(X)
            case _:
                raise ValueError(f"{selection_criterion} not implemented")

    selected_dist = min(results, key=results.get)
    return selected_dist



"""Univariate Student's t Estimation."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy.typing as npt
import scipy.stats as st

from skfolio.distribution.univariate._base import BaseUnivariateDist


class StudentT(BaseUnivariateDist):
    r"""Student's t Distribution Estimation.

    This estimator fits a univariate Student's t distribution to the input data.

    The probability density function is:

    .. math::

        f(x, \nu) = \frac{\Gamma((\nu+1)/2)}
                        {\sqrt{\pi \nu} \Gamma(\nu/2)}
                    (1+x^2/\nu)^{-(\nu+1)/2}

    where :math:`x` is a real number and the degrees of freedom parameter :math:`\nu`
    (denoted `dof` in the implementation) satisfies :math:`\nu > 0`. :math:`\Gamma` is
    the gamma function (`scipy.special.gamma`).

    The probability density above is defined in the "standardized" form. To shift
    and/or scale the distribution use the loc and scale parameters. Specifically,
    `pdf(x, df, loc, scale)` is equivalent to `pdf(y, df) / scale` with
    `y = (x - loc) / scale`.

    For more information, you can refer to the `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.t.html#scipy.stats.t>`_


    Parameters
    ----------
    loc : float or None, default=None
        If provided, the location parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    scale : float or None, default=None
        If provided, the scale parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    dof_ : float
        The fitted degrees of freedom for the Student's t distribution.

    loc_ : float
        The fitted location parameter.

    scale_ : float
        The fitted scale parameter.

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_index
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution.univariate import StudentT
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_index()
    >>> X = prices_to_returns(prices)
    >>>
    >>> # Initialize the estimator.
    >>> model = StudentT()
    >>>
    >>> # Fit the model to the data.
    >>> model.fit(X)
    >>>
    >>> # Display the fitted parameters.
    >>> print(model.fitted_repr)
    StudentT(2.75, 0.000618, 0.00681)
    >>>
    >>> # Compute the log-likelihood, total log-likelihood, CDF, PPF, AIC, and BIC
    >>> log_likelihood = model.score_samples(X)
    >>> score = model.score(X)
    >>> cdf = model.cdf(X)
    >>> ppf = model.ppf(X)
    >>> aic = model.aic(X)
    >>> bic = model.bic(X)
    >>>
    >>> # Generate 5 new samples from the fitted distribution.
    >>> samples = model.sample(n_samples=5)
    >>>
    >>> # Plot the estimated probability density function (PDF).
    >>> fig = model.plot_pdf()
    >>> fig.show()
    """

    dof_: float
    loc_: float
    scale_: float
    _scipy_model = st.t

    def __init__(
        self,
        loc: float | None = None,
        scale: float | None = None,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.loc = loc
        self.scale = scale

    @property
    def _scipy_params(self) -> dict[str, float]:
        """Dictionary of parameters to pass to the underlying SciPy distribution."""
        return {"loc": self.loc_, "scale": self.scale_, "df": self.dof_}

    def fit(self, X: npt.ArrayLike, y=None) -> "StudentT":
        """Fit the univariate Student's t distribution model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            The input data. X must contain a single column.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : StudentT
            Returns the instance itself.
        """
        X = self._validate_X(X, reset=True)

        fixed_params = {}
        if self.loc is not None:
            fixed_params["floc"] = self.loc
        if self.scale is not None:
            fixed_params["fscale"] = self.scale

        self.dof_, self.loc_, self.scale_ = self._scipy_model.fit(X, **fixed_params)
        return self



"""Distribution module."""

from skfolio.distribution._base import BaseDistribution, SelectionCriterion
from skfolio.distribution.copula import (
    BaseBivariateCopula,
    ClaytonCopula,
    CopulaRotation,
    GaussianCopula,
    GumbelCopula,
    IndependentCopula,
    JoeCopula,
    StudentTCopula,
    compute_pseudo_observations,
    empirical_tail_concentration,
    plot_tail_concentration,
    select_bivariate_copula,
)
from skfolio.distribution.multivariate import (
    BaseMultivariateDist,
    DependenceMethod,
    VineCopula,
)
from skfolio.distribution.univariate import (
    BaseUnivariateDist,
    Gaussian,
    JohnsonSU,
    NormalInverseGaussian,
    StudentT,
    select_univariate_dist,
)

__all__ = [
    "BaseBivariateCopula",
    "BaseDistribution",
    "BaseMultivariateDist",
    "BaseUnivariateDist",
    "ClaytonCopula",
    "CopulaRotation",
    "DependenceMethod",
    "Gaussian",
    "GaussianCopula",
    "GumbelCopula",
    "IndependentCopula",
    "JoeCopula",
    "JohnsonSU",
    "NormalInverseGaussian",
    "SelectionCriterion",
    "StudentT",
    "StudentTCopula",
    "VineCopula",
    "compute_pseudo_observations",
    "empirical_tail_concentration",
    "plot_tail_concentration",
    "select_bivariate_copula",
    "select_univariate_dist",
]



"""Base classes and utilities for portfolio optimization estimators.

This module defines the abstract `BaseOptimization` estimator that all
optimization algorithms in skfolio should inherit from.
"""

# Copyright (c) 2023-2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Mapping
from functools import wraps
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
import pandas as pd
import sklearn as sk
import sklearn.base as skb
from sklearn.utils.validation import check_is_fitted

import skfolio.typing as skt
from skfolio._constants import _ParamKey
from skfolio.measures import RatioMeasure
from skfolio.population import Population
from skfolio.portfolio import FailedPortfolio, Portfolio
from skfolio.prior import ReturnDistribution
from skfolio.utils.tools import input_to_array


class BaseOptimization(skb.BaseEstimator, ABC):
    """Base class for all portfolio optimizations in skfolio.

    Parameters
    ----------
    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        `transaction_costs`, `management_fees`, `previous_weights` and `risk_free_rate`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets,), optional
        Previous asset weights. Some estimators use this to compute costs or turnover.
        Additionally, when `fallback="previous_weights"`, failures will fall back to
        these weights if provided.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.
    """

    weights_: np.ndarray
    n_features_in_: int
    feature_names_in_: np.ndarray
    fallback_: BaseOptimization | Literal["previous_weights"] | None
    fallback_chain_: list[tuple[str, str]] | None
    error_: str | list[str] | None

    @abstractmethod
    def __init__(
        self,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        previous_weights: skt.MultiInput | None = None,
        raise_on_failure: bool = True,
    ):
        self.portfolio_params = portfolio_params
        self.fallback = fallback
        self.previous_weights = previous_weights
        self.raise_on_failure = raise_on_failure

    # Automatically wrap all subclasses' fit to add fallback behavior
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        original_fit = cls.__dict__.get("fit")
        if original_fit is None or getattr(original_fit, "_fallback_wrapped", False):
            return

        @wraps(original_fit)
        def _wrapped_fit(
            self, X: npt.ArrayLike, y: npt.ArrayLike | None = None, **fit_params
        ):
            self.fallback_ = None
            self.fallback_chain_ = None
            self.error_ = None

            try:
                original_fit(self, X, y, **fit_params)
            except Exception as primary_error:
                try:
                    self._run_fallback_chain(
                        X=X, y=y, primary_error=primary_error, **fit_params
                    )
                except Exception as last_error:
                    self.error_ = str(last_error)
                    if self.raise_on_failure:
                        raise
                    warnings.warn(
                        (
                            f"{self.__class__.__name__}.fit failed: {last_error}. "
                            "Because raise_on_failure=False, weights_ is set to None. "
                            "Inspect 'error_' and 'fallback_chain_' for details."
                        ),
                        stacklevel=2,
                    )
                    self.weights_ = None
            return self

        _wrapped_fit._fallback_wrapped = True
        cls.fit = _wrapped_fit

    def _run_fallback_chain(
        self,
        X: npt.ArrayLike,
        y: npt.ArrayLike | None,
        primary_error: Exception,
        **fit_params,
    ) -> None:
        """Execute the configured fallback chain after a primary `fit` failure.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Training data passed to `fit`.

        y : array-like or None
            Optional target data.

        primary_error : Exception
            The exception raised by the primary estimator.

        **fit_params : dict
            Additional keyword arguments forwarded to each fallback's `fit`.

        Raises
        ------
        Exception
            Re-raises the last encountered error if all fallbacks fail.
        """
        fallback = self.fallback

        if fallback is None:
            raise primary_error

        # Log the primary error in fallback_chain_ only when fallbacks are provided
        self.fallback_chain_ = [(str(self), str(primary_error))]

        n_assets = X.shape[1]

        if not isinstance(fallback, list | tuple):
            fallback = [fallback]

        if len(fallback) == 0:
            raise primary_error

        last_error: Exception = primary_error
        for fb in fallback:
            try:
                fb = _validate_fallback(fb)
                if fb == _ParamKey.PREVIOUS_WEIGHTS.value:
                    self._fallback_to_previous_weights_or_raise(n_assets=n_assets)
                    return

                fb_est = sk.clone(fb)

                # previous_weights are propagated to the fallbacks
                if self.previous_weights is not None:
                    if fb_est.previous_weights is not None:
                        warnings.warn(
                            (
                                "previous_weights are automatically propagated to "
                                "fallback estimators. To silence this warning, leave "
                                "the fallback's previous_weights as None."
                            ),
                            stacklevel=2,
                        )
                    fb_est.set_params(previous_weights=self.previous_weights)

                fb_est.fit(X, y, **fit_params)

                # Success: copy learned artifacts back to self
                for name in ("weights_", "n_features_in_", "feature_names_in_"):
                    setattr(self, name, getattr(fb_est, name))

                self.fallback_ = fb_est
                self.fallback_chain_.append((str(fb_est), "success"))
                return
            except Exception as err:  # try next fallback
                last_error = err
                self.fallback_chain_.append((str(fb), str(err)))
                continue

        # All fallbacks failed
        if last_error is not None:
            # Defer raising to the caller which decides based on raise_on_failure
            raise last_error
        raise RuntimeError(
            "All fallback estimators failed; inspect 'fallback_chain_' for details."
        )

    def _fallback_to_previous_weights_or_raise(self, n_assets: int) -> None:
        """Fallback to `previous_weights` or raise if unavailable/invalid.

        Parameters
        ----------
        n_assets : int
            Number of assets used to validate the shape of `previous_weights`.

        Raises
        ------
        RuntimeError
            If `previous_weights` is `None` when the fallback is requested.
        """
        try:
            if self.previous_weights is None:
                raise RuntimeError(
                    "Fallback 'previous_weights' requested, but 'previous_weights' is None. "
                    "Provide valid previous weights or remove this fallback."
                )
            self.weights_ = self._clean_previous_weights(n_assets=n_assets)
            self.fallback_ = _ParamKey.PREVIOUS_WEIGHTS.value
            self.fallback_chain_.append((_ParamKey.PREVIOUS_WEIGHTS.value, "success"))

        except Exception as error:
            self.fallback_chain_.append((_ParamKey.PREVIOUS_WEIGHTS.value, str(error)))
            raise

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike | None = None):
        pass

    def predict(self, X: npt.ArrayLike | ReturnDistribution) -> Portfolio | Population:
        """Predict the `Portfolio` or a `Population` of portfolios on `X`.

        Optimization estimators can return a 1D or a 2D array of `weights`.
        For a 1D array, the prediction is a single `Portfolio`.
        For a 2D array, the prediction is a `Population` of `Portfolio`.

        If `name` is not provided in the portfolio parameters, the estimator
        class name is used.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets) | ReturnDistribution
            Asset returns or a `ReturnDistribution` carrying returns and optional
            sample weights.

        Returns
        -------
        Portfolio | Population
            The predicted `Portfolio` or `Population` based on the fitted `weights`.
        """
        check_is_fitted(self, "weights_")

        if self.portfolio_params is None:
            ptf_kwargs = {}
        else:
            ptf_kwargs = self.portfolio_params.copy()

        # Set X and sample_weight
        if isinstance(X, ReturnDistribution):
            ptf_kwargs["sample_weight"] = X.sample_weight
            if hasattr(self, "feature_names_in_"):
                ptf_kwargs["X"] = pd.DataFrame(
                    X.returns, columns=self.feature_names_in_
                )
            else:
                ptf_kwargs["X"] = X.returns
        else:
            ptf_kwargs["X"] = X

        # Set the default portfolio parameters equal to the optimization parameters
        for param in [
            _ParamKey.TRANSACTION_COSTS.value,
            _ParamKey.MANAGEMENT_FEES.value,
            _ParamKey.PREVIOUS_WEIGHTS.value,
            _ParamKey.RISK_FREE_RATE.value,
        ]:
            if param not in ptf_kwargs and hasattr(self, param):
                ptf_kwargs[param] = getattr(self, param)

        # If 'name' is not provided in the portfolio arguments, we use the first
        # 500 characters of the optimization estimator's name
        name = ptf_kwargs.pop("name", type(self).__name__)

        # Add fallback chain
        ptf_kwargs["fallback_chain"] = self.fallback_chain_

        # If weights are None and raise_on_failure is False, we return a FailedPortfolio
        if self.weights_ is None:
            return FailedPortfolio(
                name=name, optimization_error=self.error_, **ptf_kwargs
            )

        # Optimization estimators can return a 1D or a 2D array of weights.
        # For a 1D array we return a portfolio.
        if self.weights_.ndim == 1:
            return Portfolio(weights=self.weights_, name=name, **ptf_kwargs)

        # For a 2D array we return a population of portfolios.
        n_portfolios = self.weights_.shape[0]
        population = Population([])
        for i in range(n_portfolios):
            ptf_name = f"ptf{i} - {name}"
            if np.isnan(self.weights_[i]).all():
                error = self.error_[i] if isinstance(self.error_, list) else None
                population.append(
                    FailedPortfolio(
                        name=ptf_name,
                        optimization_error=error,
                        **ptf_kwargs,
                    )
                )
            else:
                population.append(
                    Portfolio(weights=self.weights_[i], name=ptf_name, **ptf_kwargs)
                )
        return population

    def score(
        self, X: npt.ArrayLike | ReturnDistribution, y: npt.ArrayLike = None
    ) -> float:
        """Prediction score using the Sharpe Ratio.
        If the prediction is a single `Portfolio`, the score is its Sharpe Ratio.
        If the prediction is a `Population`, the score is the mean Sharpe Ratio
        across portfolios.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        score : float
            The Sharpe Ratio of the portfolio if the prediction is a single `Portfolio`
            or the mean of all the portfolios Sharpe Ratios if the prediction is a
            `Population` of `Portfolio`.
        """
        result = self.predict(X)
        if isinstance(result, Population):
            return result.measures_mean(RatioMeasure.SHARPE_RATIO)
        return result.sharpe_ratio

    def fit_predict(self, X):
        """Perform `fit` on `X` and returns the predicted `Portfolio` or
        `Population` of `Portfolio` on `X` based on the fitted `weights`.
        For factor models, use `fit(X, y)` then `predict(X)` separately.

        If fitting fails and `raise_on_failure=False`, this returns a
        `FailedPortfolio`.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        Returns
        -------
        Portfolio | Population
            The predicted `Portfolio` or `Population` based on the fitted `weights`.
        """
        return self.fit(X).predict(X)

    @property
    def needs_previous_weights(self) -> bool:
        """Whether `previous_weights` must be propagated between folds/rebalances.

        Used by `cross_val_predict` to decide whether to run sequentially and pass
        the weights from the previous rebalancing to the next. This is `True` when
        transaction costs, a maximum turnover, or a fallback depending on
        `previous_weights` are present.
        """
        if _has_transaction_cost(
            getattr(self, _ParamKey.TRANSACTION_COSTS.value, None)
        ):
            return True

        if getattr(self, "max_turnover", None) is not None:
            return True

        fallback = self.fallback
        if fallback is not None:
            if not isinstance(fallback, list | tuple):
                fallback = [fallback]

            for fb in fallback:
                fb = _validate_fallback(fb)
                if fb == _ParamKey.PREVIOUS_WEIGHTS.value or fb.needs_previous_weights:
                    return True

        return False

    def _clean_input(
        self,
        value: float | dict | npt.ArrayLike | None,
        n_assets: int,
        fill_value: Any,
        name: str,
    ) -> float | np.ndarray:
        """Convert input to a cleaned float or 1D ndarray.

        Parameters
        ----------
        value : float | dict | array-like | None
            Input value to clean.

        n_assets : int
            Number of assets. Used to verify the shape of the converted array.

        fill_value : Any
            When `value` is a dictionary, keys not present in the asset names are
            filled with `fill_value` in the converted array.

        name : str
            Name used for error messages.

        Returns
        -------
        float | ndarray of shape (n_assets,)
            The cleaned scalar or 1D array.
        """
        if value is None:
            return fill_value
        if np.isscalar(value):
            return float(value)
        return input_to_array(
            items=value,
            n_assets=n_assets,
            fill_value=fill_value,
            dim=1,
            assets_names=(
                self.feature_names_in_ if hasattr(self, "feature_names_in_") else None
            ),
            name=name,
        )

    def _clean_previous_weights(self, n_assets: int) -> np.ndarray:
        """Return validated previous weights as a 1D array of length `n_assets`.

        Converts `previous_weights` to a numpy array using `_clean_input`, accepting
        scalars, mappings keyed by asset name, or array-like inputs. Scalars are
        broadcast to all assets. Missing assets in mappings are filled with zeros.

        Parameters
        ----------
        n_assets : int
            Number of assets; used to validate shape and for broadcasting.

        Returns
        -------
        ndarray of shape (n_assets,)
            Cleaned previous weights.
        """
        previous_weights = self._clean_input(
            self.previous_weights,
            n_assets=n_assets,
            fill_value=0,
            name=_ParamKey.PREVIOUS_WEIGHTS.value,
        )
        if np.isscalar(previous_weights):
            previous_weights = np.full(n_assets, float(previous_weights))
        return previous_weights


def _has_transaction_cost(x: Any) -> bool:
    """Return True if any non-zero transaction cost is present in `x`.

    Accepts scalars, arrays, nested mappings, or structures convertible to arrays.
    Zero or empty values are treated as no cost.
    """
    if x is None:
        return False

    if isinstance(x, Mapping):
        # Empty dict -> no costs; otherwise recurse
        return any(_has_transaction_cost(v) for v in x.values())

    try:
        arr = np.asarray(x, dtype=float)
    except Exception:
        # If coercion fails, assume non-zero to be conservative
        return True

    if arr.size == 0:
        return False

    return not np.allclose(arr, 0.0, atol=1e-15, rtol=1e-18, equal_nan=False)


def _validate_fallback(
    fallback: Literal["previous_weights"] | BaseOptimization,
) -> Literal["previous_weights"] | BaseOptimization:
    """Validate the fallback specification.

    Parameters
    ----------
    fallback : BaseOptimization | "previous_weights"
        The configured fallback.

    Returns
    -------
    BaseOptimization | "previous_weights"
        The validated fallback, unchanged.

    Raises
    ------
    ValueError
        If `fallback` is a string different from `"previous_weights"`.
    TypeError
        If `fallback` is not a string and not an instance of `BaseOptimization`.
    """
    if isinstance(fallback, str):
        if fallback != _ParamKey.PREVIOUS_WEIGHTS.value:
            raise ValueError(
                f"Unsupported string fallback: {fallback!r}. Only 'previous_weights' is allowed."
            )
        return _ParamKey.PREVIOUS_WEIGHTS.value
    if not isinstance(fallback, BaseOptimization):
        raise TypeError(
            f"Fallback estimators must inherit from BaseOptimization (got {type(fallback).__name__})."
        )
    return fallback



"""Covariance module."""

from skfolio.moments.covariance._base import (
    BaseCovariance,
)
from skfolio.moments.covariance._denoise_covariance import DenoiseCovariance
from skfolio.moments.covariance._detone_covariance import DetoneCovariance
from skfolio.moments.covariance._empirical_covariance import EmpiricalCovariance
from skfolio.moments.covariance._ew_covariance import EWCovariance
from skfolio.moments.covariance._gerber_covariance import GerberCovariance
from skfolio.moments.covariance._graphical_lasso_cv import GraphicalLassoCV
from skfolio.moments.covariance._implied_covariance import ImpliedCovariance
from skfolio.moments.covariance._ledoit_wolf import LedoitWolf
from skfolio.moments.covariance._oas import OAS
from skfolio.moments.covariance._shrunk_covariance import ShrunkCovariance

__all__ = [
    "OAS",
    "BaseCovariance",
    "DenoiseCovariance",
    "DetoneCovariance",
    "EWCovariance",
    "EmpiricalCovariance",
    "GerberCovariance",
    "GraphicalLassoCV",
    "ImpliedCovariance",
    "LedoitWolf",
    "ShrunkCovariance",
]



"""Synthetic Data Prior estimator."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

import inspect

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

from skfolio.distribution import VineCopula
from skfolio.prior._base import BasePrior
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.utils.tools import check_estimator


class SyntheticData(BasePrior):
    """Synthetic Data Estimator.

    The Synthetic Data model estimates a :class:`~skfolio.prior.ReturnDistribution` by
    fitting a `distribution_estimator` and sampling new returns data from it.

    The default ``distribution_estimator`` is a Regular Vine Copula model. Other common
    choices are Generative Adversarial Networks (GANs) or Variational Autoencoders
    (VAEs).

    This class is particularly useful when the historical distribution tail dependencies
    are sparse and need extrapolation for tail optimizations or when optimizing under
    conditional or stressed scenarios.

    Parameters
    ----------
    distribution_estimator : BaseEstimator, optional
        Estimator to model the distribution of asset returns. It must inherit from
        `BaseEstimator` and implements a `sample` method. If None, the default
        `VineCopula()` model is used.

    n_samples : int, default=1000
        Number of samples to generate from the `distribution_estimator`, default is
        1000.

    sample_args : dict, optional
        Additional keyword arguments to pass to the `sample` method of the
        `distribution_estimator`.

    Attributes
    ----------
    return_distribution_ : ReturnDistribution
        Fitted :class:`~skfolio.prior.ReturnDistribution` to be used by the optimization
        estimators, containing the assets syntehtic data distribution and moments
        estimation.

    distribution_estimator_ : BaseEstimator
        The fitted distribution estimator.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Examples
    --------
    >>> import numpy as np
    >>> from skfolio.datasets import load_sp500_dataset, load_factors_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution import VineCopula
    >>> from skfolio.optimization import MeanRisk
    >>> from skfolio.prior import FactorModel, SyntheticData
    >>> from skfolio import RiskMeasure
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_dataset()
    >>> factors = load_factors_dataset()
    >>> X, y = prices_to_returns(prices, factors)
    >>>
    >>> # Instanciate the SyntheticData model and fit it
    >>> model = SyntheticData()
    >>> model.fit(X)
    >>> print(model.return_distribution_)
    >>>
    >>> # Minimum CVaR optimization on synthetic returns
    >>> model = MeanRisk(
    ...    risk_measure=RiskMeasure.CVAR,
    ...    prior_estimator=SyntheticData(
    ...        distribution_estimator=VineCopula(log_transform=True, n_jobs=-1),
    ...        n_samples=2000,
    ...    )
    ... )
    >>> model.fit(X)
    >>> print(model.weights_)
    >>>
    >>> # Minimum CVaR optimization on Stressed Factors
    >>> factor_model = FactorModel(
    ...    factor_prior_estimator=SyntheticData(
    ...        distribution_estimator=VineCopula(
    ...            central_assets=["QUAL"],
    ...            log_transform=True,
    ...            n_jobs=-1,
    ...        ),
    ...        n_samples=5000,
    ...        sample_args=dict(conditioning={"QUAL": -0.2}),
    ...    )
    ... )
    >>> model = MeanRisk(risk_measure=RiskMeasure.CVAR, prior_estimator=factor_model)
    >>> model.fit(X, y)
    >>> print(model.weights_)
    >>>
    >>> # Stress Test the Portfolio
    >>> factor_model.set_params(factor_prior_estimator__sample_args=dict(
    ...     conditioning={"QUAL": -0.5}
    ... ))
    >>> factor_model.fit(X,y)
    >>> stressed_dist = factor_model.return_distribution_
    >>> stressed_ptf = model.predict(stressed_dist)
    """

    distribution_estimator_: skb.BaseEstimator
    prior_estimator_: BasePrior
    n_features_in_: int
    feature_names_in_: np.ndarray

    def __init__(
        self,
        distribution_estimator: skb.BaseEstimator | None = None,
        n_samples: int = 1000,
        sample_args: dict | None = None,
    ):
        self.distribution_estimator = distribution_estimator
        self.n_samples = n_samples
        self.sample_args = sample_args

    def get_metadata_routing(self):
        # noinspection PyTypeChecker
        router = skm.MetadataRouter(owner=self.__class__.__name__).add(
            distance_estimator=self.distribution_estimator,
            method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
        )
        return router

    def fit(self, X: npt.ArrayLike, y=None, **fit_params) -> "SyntheticData":
        """Fit the Synthetic Data estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        **fit_params : dict
            Parameters to pass to the underlying estimators.
            Only available if `enable_metadata_routing=True`, which can be
            set by using ``sklearn.set_config(enable_metadata_routing=True)``.
            See :ref:`Metadata Routing User Guide <metadata_routing>` for
            more details.

        Returns
        -------
        self : SyntheticData
            Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        self.distribution_estimator_ = check_estimator(
            self.distribution_estimator,
            default=VineCopula(),
            check_type=skb.BaseEstimator,
        )
        _check_sample_method(self.distribution_estimator_)

        # fitting distribution estimator on prior returns
        # noinspection PyUnresolvedReferences
        self.distribution_estimator_.fit(
            X, y, **routed_params.distribution_estimator.fit
        )

        # We validate after all models have been fitted to keep feature names
        # information.
        skv.validate_data(self, X)

        # sample from the distribution estimator
        sample_args = self.sample_args if self.sample_args is not None else {}
        # noinspection PyUnresolvedReferences
        synthetic_data = self.distribution_estimator_.sample(
            n_samples=self.n_samples, **sample_args
        )

        # When performing conditional sampling, the conditioning samples are often
        # constant. To avoid null variance, we add a small white noise.
        constant_returns = np.var(synthetic_data, axis=0) < 1e-14
        if np.any(constant_returns):
            noise = 1e-6 * np.random.randn(len(synthetic_data), 1)
            synthetic_data[:, constant_returns] += noise

        # Fit empirical posterior estimator
        posterior_estimator = EmpiricalPrior()
        posterior_estimator.fit(synthetic_data)
        self.return_distribution_ = posterior_estimator.return_distribution_

        return self


def _check_sample_method(distribution_estimator: skb.BaseEstimator) -> None:
    """Check that the distribution_estimator implements a valid 'sample' method.

    This helper function verifies that the given estimator has a callable 'sample'
    method and that this method accepts an 'n_samples' parameter.

    Parameters
    ----------
    distribution_estimator : BaseEstimator
        The estimator whose 'sample' method is to be validated.

    Raises
    ------
    ValueError
        If the 'sample' method is missing or does not have an 'n_samples' parameter.
    """
    # Get the 'sample' attribute; if it doesn't exist, return False.
    sample_method = getattr(distribution_estimator, "sample", None)
    if sample_method is None or not callable(sample_method):
        raise ValueError(
            f"The distribution_estimator {distribution_estimator} must implement a "
            "`sample` method"
        )

    sig = inspect.signature(sample_method)

    # Check if the parameter 'n_samples' is in the method's parameters.
    if "n_samples" not in sig.parameters:
        raise ValueError(
            "The `sample` method of the distribution_estimator "
            f"{distribution_estimator} must have `n_samples` as parameter"
        )
